"""Methods for numba-accelerated bilinear and trilinear interpolation."""

import numpy as np
from numba import njit, prange

PARALLEL_MODE = True


def _is_linearly_spaced(array, tol=1e-8):
    """
    Check if a numpy array is linearly spaced within some numerical precision.

    Parameters
    ----------
    array : np.ndarray
        The input array to check.
    tol : float
        The numerical tolerance for checking linear spacing.

    Returns
    -------
    bool
        True if the array is linearly spaced within the given tolerance, False otherwise.
    """
    if len(array) < 2:
        return True

    # Calculate the differences between consecutive elements
    diffs = np.diff(array)

    # Check if all differences are approximately equal within the given tolerance
    return np.all(np.abs(diffs - diffs[0]) <= tol)


@njit(parallel=PARALLEL_MODE)
def _fast_linear_interpolate(desired_pos, orig_coords, orig_values):
    """
    Perform numba-accelerated linear interpolation on a 1D array of values.

    Parameters
    ----------
    desired_pos : np.ndarray
        The desired positions for interpolation.
    orig_coords : np.ndarray
        The original coordinates. These should be monotonic but need not be linearly spaced.
    orig_values : np.ndarray
        The values at the original coordinates.

    Returns
    -------
    np.ndarray
        The interpolated values at the desired positions.
    """
    # Check if the original coordinates are decreasing and reverse them if necessary
    if orig_coords[0] > orig_coords[-1]:
        orig_coords = orig_coords[::-1]
        orig_values = orig_values[::-1]

    # Flatten the desired positions
    desired_pos = desired_pos.flatten()
    n_points = desired_pos.size
    result = np.empty(n_points)

    for idx in prange(n_points):
        x = desired_pos[idx]

        # Find the indices of the grid points surrounding x
        x1_idx = np.searchsorted(orig_coords, x) - 1
        x2_idx = x1_idx + 1

        # Boundary check to ensure we do not go out of bounds
        if x1_idx < 0 or x2_idx >= len(orig_coords):
            result[idx] = np.nan
            continue

        # Coordinates for surrounding points
        x1 = orig_coords[x1_idx]
        x2 = orig_coords[x2_idx]

        # Values at surrounding points
        Q1 = orig_values[x1_idx]
        Q2 = orig_values[x2_idx]

        # Perform linear interpolation
        result[idx] = (Q1 * (x2 - x) + Q2 * (x - x1)) / (x2 - x1)

    return result


@njit(parallel=PARALLEL_MODE)
def _fast_linear_interpolate_rectilinear(desired_pos, orig_coords, orig_values):
    """
    Perform numba-accelerated linear interpolation on a 1D array of values assuming a linearly spaced input grid.

    Parameters
    ----------
    desired_pos : np.ndarray
        The desired positions for interpolation.
    orig_coords : np.ndarray
        The original coordinates. These must be linearly spaced.
    orig_values : np.ndarray
        The values at the original coordinates.

    Returns
    -------
    np.ndarray
        The interpolated values at the desired positions.
    """
    # Check if the original coordinates are decreasing and reverse them if necessary
    if orig_coords[0] > orig_coords[-1]:
        orig_coords = orig_coords[::-1]
        orig_values = orig_values[::-1]

    # Flatten the desired positions
    desired_pos = desired_pos.flatten()
    n_points = desired_pos.size
    result = np.empty(n_points)

    # Calculate the step size
    step = (orig_coords[-1] - orig_coords[0]) / (len(orig_coords) - 1)

    for idx in prange(n_points):
        x = desired_pos[idx]

        # Calculate the indices of the grid points surrounding x
        x1_idx = int((x - orig_coords[0]) / step)
        x2_idx = x1_idx + 1

        # Boundary check to ensure we do not go out of bounds
        if x1_idx < 0 or x2_idx >= len(orig_coords):
            result[idx] = np.nan
            continue

        # Coordinates for surrounding points
        x1 = orig_coords[x1_idx]
        x2 = orig_coords[x2_idx]

        # Values at surrounding points
        Q1 = orig_values[x1_idx]
        Q2 = orig_values[x2_idx]

        # Perform linear interpolation
        result[idx] = (Q1 * (x2 - x) + Q2 * (x - x1)) / (x2 - x1)

    return result


@njit(parallel=PARALLEL_MODE)
def _fast_bilinear_interpolate(
    desired_pos_dim0,
    desired_pos_dim1,
    orig_coords_dim0,
    orig_coords_dim1,
    orig_values,
):
    """
    Perform numba-accelerated bilinear interpolation on a 2D grid of values.

    Parameters
    ----------
    desired_pos_dim0 : np.ndarray
        The desired positions along the first dimension.
    desired_pos_dim1 : np.ndarray
        The desired positions along the second dimension.
        Should have the same shape as `desired_pos_dim0`.
    orig_coords_dim0 : np.ndarray
        The original coordinates along the first dimension.
        These should be monotonic but need not be linearly spaced.
    orig_coords_dim1 : np.ndarray
        The original coordinates along the second dimension.
        These should be monotonic but need not be linearly spaced.
    orig_values : np.ndarray
        The values at the original grid points.

    Returns
    -------
    np.ndarray
        The interpolated values at the desired positions.
    """
    # Check if the original coordinates are decreasing and reverse them if necessary
    if orig_coords_dim0[0] > orig_coords_dim0[-1]:
        orig_coords_dim0 = orig_coords_dim0[::-1]
        orig_values = orig_values[::-1, :]
    if orig_coords_dim1[0] > orig_coords_dim1[-1]:
        orig_coords_dim1 = orig_coords_dim1[::-1]
        orig_values = orig_values[:, ::-1]

    # Flatten the desired positions
    desired_shape = desired_pos_dim0.shape  # Store for later
    desired_pos_dim0 = desired_pos_dim0.flatten()
    desired_pos_dim1 = desired_pos_dim1.flatten()

    n_points = desired_pos_dim0.size
    result = np.empty(n_points)

    for idx in prange(n_points):
        x = desired_pos_dim0[idx]
        y = desired_pos_dim1[idx]

        # Find the indices of the grid points surrounding (x, y)
        x1_idx = np.searchsorted(orig_coords_dim0, x) - 1
        x2_idx = x1_idx + 1
        y1_idx = np.searchsorted(orig_coords_dim1, y) - 1
        y2_idx = y1_idx + 1

        # Boundary check to ensure we do not go out of bounds
        if (
            x1_idx < 0
            or x2_idx >= len(orig_coords_dim0)
            or y1_idx < 0
            or y2_idx >= len(orig_coords_dim1)
        ):
            result[idx] = np.nan
            continue

        # Coordinates for surrounding points
        x1 = orig_coords_dim0[x1_idx]
        x2 = orig_coords_dim0[x2_idx]
        y1 = orig_coords_dim1[y1_idx]
        y2 = orig_coords_dim1[y2_idx]

        # Values at surrounding points
        Q11 = orig_values[x1_idx, y1_idx]
        Q12 = orig_values[x1_idx, y2_idx]
        Q21 = orig_values[x2_idx, y1_idx]
        Q22 = orig_values[x2_idx, y2_idx]

        # Perform bilinear interpolation
        result[idx] = (
            Q11 * (x2 - x) * (y2 - y)
            + Q21 * (x - x1) * (y2 - y)
            + Q12 * (x2 - x) * (y - y1)
            + Q22 * (x - x1) * (y - y1)
        ) / ((x2 - x1) * (y2 - y1))

    return result.reshape(desired_shape)


@njit(parallel=PARALLEL_MODE)
def _fast_bilinear_interpolate_rectilinear(
    desired_pos_dim0,
    desired_pos_dim1,
    orig_coords_dim0,
    orig_coords_dim1,
    orig_values,
):
    """
    Perform numba-accelerated bilinear interpolation on a rectilinear 2D grid of values.

    Parameters
    ----------
    desired_pos_dim0 : np.ndarray
        The desired positions along the first dimension.
    desired_pos_dim1 : np.ndarray
        The desired positions along the second dimension.
        Should have the same shape as `desired_pos_dim0`.
    orig_coords_dim0 : np.ndarray
        The original coordinates along the first dimension.
        These must be linearly spaced.
    orig_coords_dim1 : np.ndarray
        The original coordinates along the second dimension.
        These must be linearly spaced.
    orig_values : np.ndarray
        The values at the original grid points.

    Returns
    -------
    np.ndarray
        The interpolated values at the desired positions.
    """
    # Check if the original coordinates are decreasing and reverse them if necessary
    if orig_coords_dim0[0] > orig_coords_dim0[-1]:
        orig_coords_dim0 = orig_coords_dim0[::-1]
        orig_values = orig_values[::-1, :]
    if orig_coords_dim1[0] > orig_coords_dim1[-1]:
        orig_coords_dim1 = orig_coords_dim1[::-1]
        orig_values = orig_values[:, ::-1]

    # Flatten the desired positions
    desired_shape = desired_pos_dim0.shape  # Store for later
    desired_pos_dim0 = desired_pos_dim0.flatten()
    desired_pos_dim1 = desired_pos_dim1.flatten()

    n_points = desired_pos_dim0.size
    result = np.empty(n_points)

    # Calculate the step sizes
    step_dim0 = (orig_coords_dim0[-1] - orig_coords_dim0[0]) / (
        len(orig_coords_dim0) - 1
    )
    step_dim1 = (orig_coords_dim1[-1] - orig_coords_dim1[0]) / (
        len(orig_coords_dim1) - 1
    )

    for idx in prange(n_points):
        x = desired_pos_dim0[idx]
        y = desired_pos_dim1[idx]

        # Calculate the indices of the grid points surrounding (x, y)
        x1_idx = int((x - orig_coords_dim0[0]) / step_dim0)
        x2_idx = x1_idx + 1
        y1_idx = int((y - orig_coords_dim1[0]) / step_dim1)
        y2_idx = y1_idx + 1

        # Boundary check to ensure we do not go out of bounds
        if (
            x1_idx < 0
            or x2_idx >= len(orig_coords_dim0)
            or y1_idx < 0
            or y2_idx >= len(orig_coords_dim1)
        ):
            result[idx] = np.nan
            continue

        # Coordinates for surrounding points
        x1 = orig_coords_dim0[x1_idx]
        x2 = orig_coords_dim0[x2_idx]
        y1 = orig_coords_dim1[y1_idx]
        y2 = orig_coords_dim1[y2_idx]

        # Values at surrounding points
        Q11 = orig_values[x1_idx, y1_idx]
        Q12 = orig_values[x1_idx, y2_idx]
        Q21 = orig_values[x2_idx, y1_idx]
        Q22 = orig_values[x2_idx, y2_idx]

        # Perform bilinear interpolation
        result[idx] = (
            Q11 * (x2 - x) * (y2 - y)
            + Q21 * (x - x1) * (y2 - y)
            + Q12 * (x2 - x) * (y - y1)
            + Q22 * (x - x1) * (y - y1)
        ) / ((x2 - x1) * (y2 - y1))

    return result.reshape(desired_shape)


@njit(parallel=PARALLEL_MODE)
def _fast_trilinear_interpolate(
    desired_pos_dim0,
    desired_pos_dim1,
    desired_pos_dim2,
    orig_coords_dim0,
    orig_coords_dim1,
    orig_coords_dim2,
    orig_values,
    progress_proxy=None,
):
    """
    Perform numba-accelerated trilinear interpolation on a 3D grid of values.

    Parameters
    ----------
    desired_pos_dim0 : np.ndarray
        The desired positions along the first dimension.
    desired_pos_dim1 : np.ndarray
        The desired positions along the second dimension.
        Should have the same shape as `desired_pos_dim0`.
    desired_pos_dim2 : np.ndarray
        The desired positions along the third dimension.
        Should have the same shape as `desired_pos_dim0`.
    orig_coords_dim0 : np.ndarray
        The original coordinates along the first dimension.
        These should be monotonic but need not be linearly spaced.
    orig_coords_dim1 : np.ndarray
        The original coordinates along the second dimension.
        These should be monotonic but need not be linearly spaced.
    orig_coords_dim2 : np.ndarray
        The original coordinates along the third dimension.
        These should be monotonic but need not be linearly spaced.
    orig_values : np.ndarray
        The values at the original grid points.
    progress_proxy : ProgressProxy, optional
        A numba-progress ProgressBar proxy to update the progress of the interpolation.

    Returns
    -------
    np.ndarray
      The interpolated values at the desired positions.
    """
    # Check if the original coordinates are decreasing and reverse them if necessary
    if orig_coords_dim0[0] > orig_coords_dim0[-1]:
        orig_coords_dim0 = orig_coords_dim0[::-1]
        orig_values = orig_values[::-1, :, :]
    if orig_coords_dim1[0] > orig_coords_dim1[-1]:
        orig_coords_dim1 = orig_coords_dim1[::-1]
        orig_values = orig_values[:, ::-1, :]
    if orig_coords_dim2[0] > orig_coords_dim2[-1]:
        orig_coords_dim2 = orig_coords_dim2[::-1]
        orig_values = orig_values[:, :, ::-1]

    # Flatten the desired positions
    desired_shape = desired_pos_dim0.shape  # Store for later
    desired_pos_dim0 = desired_pos_dim0.flatten()
    desired_pos_dim1 = desired_pos_dim1.flatten()
    desired_pos_dim2 = desired_pos_dim2.flatten()

    n_points = desired_pos_dim0.size
    result = np.empty(n_points)

    for idx in prange(n_points):
        if progress_proxy is not None and (idx % 100 == 0 or idx == n_points - 1):
            progress_proxy.update(100)

        x = desired_pos_dim0[idx]
        y = desired_pos_dim1[idx]
        z = desired_pos_dim2[idx]

        # Find the indices of the grid points surrounding (x, y, z)
        x1_idx = np.searchsorted(orig_coords_dim0, x) - 1
        x2_idx = x1_idx + 1
        y1_idx = np.searchsorted(orig_coords_dim1, y) - 1
        y2_idx = y1_idx + 1
        z1_idx = np.searchsorted(orig_coords_dim2, z) - 1
        z2_idx = z1_idx + 1

        # Boundary check to ensure we do not go out of bounds
        if (
            x1_idx < 0
            or x2_idx >= len(orig_coords_dim0)
            or y1_idx < 0
            or y2_idx >= len(orig_coords_dim1)
            or z1_idx < 0
            or z2_idx >= len(orig_coords_dim2)
        ):
            result[idx] = np.nan
            continue

        # Coordinates for surrounding points
        x1 = orig_coords_dim0[x1_idx]
        x2 = orig_coords_dim0[x2_idx]
        y1 = orig_coords_dim1[y1_idx]
        y2 = orig_coords_dim1[y2_idx]
        z1 = orig_coords_dim2[z1_idx]
        z2 = orig_coords_dim2[z2_idx]

        # Values at surrounding points
        Q111 = orig_values[x1_idx, y1_idx, z1_idx]
        Q112 = orig_values[x1_idx, y1_idx, z2_idx]
        Q121 = orig_values[x1_idx, y2_idx, z1_idx]
        Q122 = orig_values[x1_idx, y2_idx, z2_idx]
        Q211 = orig_values[x2_idx, y1_idx, z1_idx]
        Q212 = orig_values[x2_idx, y1_idx, z2_idx]
        Q221 = orig_values[x2_idx, y2_idx, z1_idx]
        Q222 = orig_values[x2_idx, y2_idx, z2_idx]

        # Perform trilinear interpolation
        result[idx] = (
            Q111 * (x2 - x) * (y2 - y) * (z2 - z)
            + Q211 * (x - x1) * (y2 - y) * (z2 - z)
            + Q121 * (x2 - x) * (y - y1) * (z2 - z)
            + Q221 * (x - x1) * (y - y1) * (z2 - z)
            + Q112 * (x2 - x) * (y2 - y) * (z - z1)
            + Q212 * (x - x1) * (y2 - y) * (z - z1)
            + Q122 * (x2 - x) * (y - y1) * (z - z1)
            + Q222 * (x - x1) * (y - y1) * (z - z1)
        ) / ((x2 - x1) * (y2 - y1) * (z2 - z1))

    return result.reshape(desired_shape)


@njit(parallel=PARALLEL_MODE)
def _fast_trilinear_interpolate_rectilinear(
    desired_pos_dim0,
    desired_pos_dim1,
    desired_pos_dim2,
    orig_coords_dim0,
    orig_coords_dim1,
    orig_coords_dim2,
    orig_values,
    progress_proxy=None,
):
    """
    Perform numba-accelerated trilinear interpolation on a rectilinear 3D grid of values.

    Parameters
    ----------
    desired_pos_dim0 : np.ndarray
        The desired positions along the first dimension.
    desired_pos_dim1 : np.ndarray
        The desired positions along the second dimension.
        Should have the same shape as `desired_pos_dim0`.
    desired_pos_dim2 : np.ndarray
        The desired positions along the third dimension.
        Should have the same shape as `desired_pos_dim0`.
    orig_coords_dim0 : np.ndarray
        The original coordinates along the first dimension.
        These must be linearly spaced.
    orig_coords_dim1 : np.ndarray
        The original coordinates along the second dimension.
        These must be linearly spaced.
    orig_coords_dim2 : np.ndarray
        The original coordinates along the third dimension.
        These must be linearly spaced.
    orig_values : np.ndarray
        The values at the original grid points.
    progress_proxy : ProgressProxy, optional
        A numba-progress ProgressBar proxy to update the progress of the interpolation.

    Returns
    -------
    np.ndarray
        The interpolated values at the desired positions.
    """
    # Check if the original coordinates are decreasing and reverse them if necessary
    if orig_coords_dim0[0] > orig_coords_dim0[-1]:
        orig_coords_dim0 = orig_coords_dim0[::-1]
        orig_values = orig_values[::-1, :, :]
    if orig_coords_dim1[0] > orig_coords_dim1[-1]:
        orig_coords_dim1 = orig_coords_dim1[::-1]
        orig_values = orig_values[:, ::-1, :]
    if orig_coords_dim2[0] > orig_coords_dim2[-1]:
        orig_coords_dim2 = orig_coords_dim2[::-1]
        orig_values = orig_values[:, :, ::-1]

    # Flatten the desired positions
    desired_shape = desired_pos_dim0.shape  # Store for later
    desired_pos_dim0 = desired_pos_dim0.flatten()
    desired_pos_dim1 = desired_pos_dim1.flatten()
    desired_pos_dim2 = desired_pos_dim2.flatten()

    n_points = desired_pos_dim0.size
    result = np.empty(n_points)

    # Calculate the step sizes
    step_dim0 = (orig_coords_dim0[-1] - orig_coords_dim0[0]) / (
        len(orig_coords_dim0) - 1
    )
    step_dim1 = (orig_coords_dim1[-1] - orig_coords_dim1[0]) / (
        len(orig_coords_dim1) - 1
    )
    step_dim2 = (orig_coords_dim2[-1] - orig_coords_dim2[0]) / (
        len(orig_coords_dim2) - 1
    )

    for idx in prange(n_points):
        if progress_proxy is not None and (idx % 100 == 0 or idx == n_points - 1):
            progress_proxy.update(100)

        x = desired_pos_dim0[idx]
        y = desired_pos_dim1[idx]
        z = desired_pos_dim2[idx]

        # Calculate the indices of the grid points surrounding (x, y, z)
        x1_idx = int((x - orig_coords_dim0[0]) / step_dim0)
        x2_idx = x1_idx + 1
        y1_idx = int((y - orig_coords_dim1[0]) / step_dim1)
        y2_idx = y1_idx + 1
        z1_idx = int((z - orig_coords_dim2[0]) / step_dim2)
        z2_idx = z1_idx + 1

        # Boundary check to ensure we do not go out of bounds
        if (
            x1_idx < 0
            or x2_idx >= len(orig_coords_dim0)
            or y1_idx < 0
            or y2_idx >= len(orig_coords_dim1)
            or z1_idx < 0
            or z2_idx >= len(orig_coords_dim2)
        ):
            result[idx] = np.nan
            continue

        # Coordinates for surrounding points
        x1 = orig_coords_dim0[x1_idx]
        x2 = orig_coords_dim0[x2_idx]
        y1 = orig_coords_dim1[y1_idx]
        y2 = orig_coords_dim1[y2_idx]
        z1 = orig_coords_dim2[z1_idx]
        z2 = orig_coords_dim2[z2_idx]

        # Values at surrounding points
        Q111 = orig_values[x1_idx, y1_idx, z1_idx]
        Q112 = orig_values[x1_idx, y1_idx, z2_idx]
        Q121 = orig_values[x1_idx, y2_idx, z1_idx]
        Q122 = orig_values[x1_idx, y2_idx, z2_idx]
        Q211 = orig_values[x2_idx, y1_idx, z1_idx]
        Q212 = orig_values[x2_idx, y1_idx, z2_idx]
        Q221 = orig_values[x2_idx, y2_idx, z1_idx]
        Q222 = orig_values[x2_idx, y2_idx, z2_idx]

        # Perform trilinear interpolation
        result[idx] = (
            Q111 * (x2 - x) * (y2 - y) * (z2 - z)
            + Q211 * (x - x1) * (y2 - y) * (z2 - z)
            + Q121 * (x2 - x) * (y - y1) * (z2 - z)
            + Q221 * (x - x1) * (y - y1) * (z2 - z)
            + Q112 * (x2 - x) * (y2 - y) * (z - z1)
            + Q212 * (x - x1) * (y2 - y) * (z - z1)
            + Q122 * (x2 - x) * (y - y1) * (z - z1)
            + Q222 * (x - x1) * (y - y1) * (z - z1)
        ) / ((x2 - x1) * (y2 - y1) * (z2 - z1))

    return result.reshape(desired_shape)



"""Thin wrapper around hvplot to set some defaults and automate handling pint dequantify"""

import xarray as xr


@xr.register_dataarray_accessor("iplot")
class HVPlotAccessor:
    """Thin wrapper around hvplot to handle default options and pint dequantify."""

    def __init__(self, xarray_obj):
        import hvplot.xarray  # noqa: F401

        self._obj = xarray_obj

    def __call__(self, *args, **kwargs):
        # Generate the plot
        plot = self._obj.pint.dequantify().hvplot(*args, **kwargs)
        return plot



"""Functions used to apply k-space conversions to data."""

import numba_progress
import numexpr as ne
import numpy as np
import pint
import pint_xarray
import xarray as xr
from scipy.constants import angstrom, electron_volt, hbar, m_e
from tqdm.notebook import tqdm

from peaks.core.fileIO.base_data_classes.base_data_class import BaseDataLoader
from peaks.core.process.fermi_level_correction import (
    _get_BE_scale,
    _get_E_shift_at_theta_par,
    _get_wf,
)
from peaks.core.utils.interpolation import (
    _fast_bilinear_interpolate,
    _fast_bilinear_interpolate_rectilinear,
    _fast_trilinear_interpolate,
    _fast_trilinear_interpolate_rectilinear,
    _is_linearly_spaced,
)
from peaks.core.utils.misc import analysis_warning

ureg = pint_xarray.unit_registry

# Calculate kvac_const
KVAC_CONST = (2 * m_e / (hbar**2)) ** 0.5 * (electron_volt**0.5) * angstrom
PI = np.pi


# --------------------------------------------------------- #
# Mapping functions: angle -> k-space (in plane)            #
# following conventions and nomenclature of Ishida and Shin #
# Rev. Sci. Instrum. 89 (2018) 043903.                      #
# --------------------------------------------------------- #


def ensure_radians(func):
    def wrapper(*args, **kwargs):
        processed_args = []
        for arg in args:
            if isinstance(arg, pint.Quantity):
                if arg.dimensionality != ureg.radian.dimensionality:
                    raise ValueError("All angles must be in radians.")
                else:
                    processed_args.append(arg.magnitude)
            else:
                processed_args.append(arg)

        processed_kwargs = {}
        for key, arg in kwargs.items():
            if isinstance(arg, pint.Quantity):
                if arg.dimensionality != ureg.radian.dimensionality:
                    raise ValueError("All angles must be in radians.")
                else:
                    processed_kwargs[key] = arg.magnitude
            else:
                processed_kwargs[key] = arg

        return func(*processed_args, **processed_kwargs)

    return wrapper


def _f_dispatcher(
    ana_type, Ek, alpha, beta, beta_0, chi, chi_0, delta, delta_0, xi, xi_0
):
    if ana_type == "I":
        return _fI(alpha, beta - beta_0, delta - delta_0, xi - xi_0, Ek)
    elif ana_type == "II":
        return _fII(alpha, beta - beta_0, delta - delta_0, xi - xi_0, Ek)
    elif ana_type == "Ip":
        return _fIp(alpha, beta, delta - delta_0, xi - xi_0, chi - chi_0, Ek)
    elif ana_type == "IIp":
        return _fIIp(alpha, beta, delta - delta_0, xi - xi_0, chi - chi_0, Ek)


def _f_inv_dispatcher(
    ana_type, Ek, kx, ky, beta_0, chi, chi_0, delta, delta_0, xi, xi_0
):
    if ana_type == "I":
        return _fI_inv(kx, ky, delta - delta_0, xi - xi_0, Ek, beta_0)
    elif ana_type == "II":
        return _fII_inv(kx, ky, delta - delta_0, xi - xi_0, Ek, beta_0)
    elif ana_type == "Ip":
        return _fIp_inv(kx, ky, delta - delta_0, xi - xi_0, chi - chi_0, Ek)
    elif ana_type == "IIp":
        return _fIIp_inv(kx, ky, delta - delta_0, xi - xi_0, chi - chi_0, Ek)


def _fI(alpha, beta_, delta_, xi_, Ek):  # Type I, no deflector
    """Convert angles to k-space for a type-I analyser with no deflector,
    following the conventions and nomenclature of Ishida and Shin, Rev. Sci. Instrum. 89 (2018) 043903.

    Parameters
    -----------
    alpha : float or np.ndarray
        theta_par angle (analyser slit angle, rad)
    beta_ : float or np.ndaarray
        polar angle - reference angle (rad)
    delta_ : float
        azi angle - reference angle (rad)
    xi_ : float or np.ndarray
        tilt angle - reference angle (rad)
    Ek : float or np.nadarray
        Kinetic energy (eV)

    Returns
    --------
    kx : float or np.ndarray
        k-vector along analyser slit (1/A)
    ky : float or np.ndarray
        k-vector perp to analyser slit (1/A)
    """

    # k_vacuum from KE
    kvac_str = "KVAC_CONST * sqrt(Ek)"

    # Mapping functions
    kx = ne.evaluate(
        f"{kvac_str} * (((sin(delta_) * sin(beta_) + cos(delta_) * sin(xi_) * cos(beta_)) * cos(alpha)) "
        "- (cos(delta_) * cos(xi_) * sin(alpha)))"
    )
    ky = ne.evaluate(
        f"{kvac_str} * (((-cos(delta_) * sin(beta_) + sin(delta_) * sin(xi_) * cos(beta_)) * cos(alpha)) "
        "- (sin(delta_) * cos(xi_) * sin(alpha)))"
    )

    return kx, ky


def _fII(alpha, beta_, delta_, xi_, Ek):  # Type II, no deflector
    """Convert angles to k-space for a type-II analyser with no deflector,
    following the conventions and nomenclature of Ishida and Shin, Rev. Sci. Instrum. 89 (2018) 043903.

       Parameters
       ----------
       alpha : float or np.ndarray
           theta_par angle (analyser slit angle, rad)
       beta_ : float or np.ndarray
           tilt angle - reference angle (rad)
       delta_ : float
           azi angle - reference angle (rad)
       xi_ : float or np.ndarray
           polar angle - reference angle (rad)
       Ek : float or np.ndarray
           Kinetic energy (eV)

       Returns
       -------
       kx : float or np.ndarray
           k-vector perp to analyser slit (1/A)
       ky : float or np.ndarray
           k-vector along analyser slit (1/A)
    """
    # k_vacuum from KE
    kvac_str = "KVAC_CONST * sqrt(Ek)"

    # Mapping functions
    kx = ne.evaluate(
        f"{kvac_str} * (((sin(delta_) * sin(xi_)) + (cos(delta_) * sin(beta_) * cos(xi_))) * cos(alpha) - "
        "((sin(delta_) * cos(xi_) - (cos(delta_) * sin(beta_) * sin(xi_))) * sin(alpha)))"
    )
    ky = ne.evaluate(
        f"{kvac_str} * (((-cos(delta_) * sin(xi_)) + (sin(delta_) * sin(beta_) * cos(xi_))) * cos(alpha) + "
        "((cos(delta_) * cos(xi_) + (sin(delta_) * sin(beta_) * sin(xi_))) * sin(alpha)))"
    )

    return kx, ky


def _fIp(alpha, beta, delta_, xi_, chi_, Ek):  # Type I with deflector
    """
    Convert angles to k-space for a type-I analyser with deflector, following the conventions and nomenclature of Ishida and Shin, Rev. Sci. Instrum. 89 (2018) 043903.

    Parameters
    ----------
    alpha : float or np.ndarray
        theta_par angle (analyser slit angle) + defl_par (deflector angle along the slit) (rad)
    beta : float or np.ndarray
        defl_perp (deflector angle perpendicular to the slit, rad)
    delta_ : float
        azi angle - reference angle (rad)
    xi_ : float or np.ndarray
        tilt angle - reference angle (rad)
    chi_ : float or np.ndarray
        polar angle - reference angle (rad)
    Ek : float or np.ndarray
        Kinetic energy (eV)

    Returns
    -------
    kx : float or np.ndarray
        k-vector along analyser slit (1/A)
    ky : float or np.ndarray
        k-vector perp to analyser slit (1/A)
    """

    # k_vacuum from KE
    kvac_str = "KVAC_CONST * sqrt(Ek)"

    # Mapping functions
    sinc_a2b2_str = "sin(sqrt(alpha**2 + beta**2))/(sqrt(alpha**2 + beta**2))"
    kx = ne.evaluate(
        f"{kvac_str} * ("
        "(((-alpha * cos(delta_) * cos(xi_)) + (beta * sin(delta_) * cos(chi_)) "
        f"- (beta * cos(delta_) * sin(xi_) * sin(chi_))) * {sinc_a2b2_str}) + "
        "(((sin(delta_) * sin(chi_)) + (cos(delta_) * sin(xi_) * cos(chi_))) * cos(sqrt(alpha**2 + beta**2)))"
        ")"
    )
    ky = ne.evaluate(
        f"{kvac_str} * ("
        "(((-alpha * sin(delta_) * cos(xi_)) - (beta * cos(delta_) * cos(chi_)) "
        f"- (beta * sin(delta_) * sin(xi_) * sin(chi_))) * {sinc_a2b2_str}) - "
        "(((cos(delta_) * sin(chi_)) - (sin(delta_) * sin(xi_) * cos(chi_))) * cos(sqrt(alpha**2 + beta**2)))"
        ")"
    )

    return kx, ky


def _fIIp(alpha, beta, delta_, xi_, chi_, Ek):  # Type II with deflector
    """
    Convert angles to k-space for a type-II analyser with deflector, following the conventions and nomenclature of Ishida and Shin, Rev. Sci. Instrum. 89 (2018) 043903.

    Parameters
    ----------
    alpha : float or np.ndarray
        theta_par angle (analyser slit angle) + defl_par (deflector angle along the slit) (rad)
    beta : float or np.ndarray
        defl_perp (deflector angle perpendicular to the slit, rad)
    delta_ : float
        azi angle - reference angle (rad)
    xi_ : float or np.ndarray
        tilt angle - reference angle (rad)
    chi_ : float or np.ndarray
        polar angle - reference angle (rad)
    Ek : float or np.ndarray
        Kinetic energy (eV)

    Returns
    -------
    kx : float or np.ndarray
        k-vector perp to analyser slit (1/A)
    ky : float or np.ndarray
        k-vector along analyser slit (1/A)
    """

    # k_vacuum from KE
    kvac_str = "KVAC_CONST * sqrt(Ek)"

    # Mapping function
    sinc_a2b2_str = "sin(sqrt(alpha**2 + beta**2))/(sqrt(alpha**2 + beta**2))"
    kx = ne.evaluate(
        f"{kvac_str} * ("
        "(((-beta * cos(delta_) * cos(xi_)) - (alpha * sin(delta_) * cos(chi_)) "
        f"+ (alpha * cos(delta_) * sin(xi_) * sin(chi_))) * {sinc_a2b2_str}) + "
        "(((sin(delta_) * sin(chi_)) + (cos(delta_) * sin(xi_) * cos(chi_))) * cos(sqrt(alpha**2 + beta**2)))"
        ")"
    )
    ky = ne.evaluate(
        f"{kvac_str} * ("
        "(((-beta * sin(delta_) * cos(xi_)) + (alpha * cos(delta_) * cos(chi_)) "
        f"+ (alpha * sin(delta_) * sin(xi_) * sin(chi_))) * {sinc_a2b2_str}) - "
        "(((cos(delta_) * sin(chi_)) - (sin(delta_) * sin(xi_) * cos(chi_))) * cos(sqrt(alpha**2 + beta**2)))"
        ")"
    )

    return kx, ky


# Inverse mapping functions for converting between angle and k-space: k --> angle


def _fI_inv(kx, ky, delta_, xi, Ek, beta_0):  # Type I, no deflector
    """
    Convert k-space to angles for a type-I analyser with no deflector, following the conventions and nomenclature
    of Ishida and Shin, Rev. Sci. Instrum. 89 (2018) 043903.

    Parameters
    ----------
    kx : float or np.ndarray
        k-vector along analyser slit (1/A)
    ky : float or np.ndarray
        k-vector perp to analyser slit (1/A)
    delta_ : float
        azi angle - reference angle (rad)
    xi : float
        tilt angle (rad)
    Ek : float
        Kinetic energy (eV)
    beta_0 : float, optional
        polar angle offset

    Returns
    -------
    alpha : float or np.ndarray
        theta_par angle (analyser slit angle, rad)
    beta : float or np.ndarray
        polar angle (rad)
    """
    # k_vacuum from KE
    kvac_str = "(KVAC_CONST * sqrt(Ek))"

    # Mapping function (include convert to degrees directly for speed)
    alpha = ne.evaluate(
        f"arcsin((sin(xi) * sqrt({kvac_str}**2 - kx**2 - ky**2) - cos(xi) * (kx * cos(delta_) + ky * sin(delta_))) "
        f"/ {kvac_str})"
    )

    beta = ne.evaluate(
        "(beta_0 + (arctan((kx * sin(delta_) - ky * cos(delta_)) /"
        " (kx * sin(xi) * cos(delta_) + ky * sin(xi) * sin(delta_) "
        f"+ cos(xi) * sqrt({kvac_str}**2 - kx**2 - ky**2)))))"
    )

    return alpha, beta


def _fII_inv(kx, ky, delta_, xi, Ek, beta_0):  # Type I, no deflector
    """
    Convert k-space to angles for a type-II analyser with no deflector, following the conventions and nomenclature of Ishida and Shin, Rev. Sci. Instrum. 89 (2018) 043903.

    Parameters
    ----------
    kx : float or np.ndarray
        k-vector perp to analyser slit (1/A)
    ky : float or np.ndarray
        k-vector along analyser slit (1/A)
    delta_ : float
        azi angle - reference angle (rad)
    xi : float
        polar angle (rad)
    Ek : float or np.ndarray
        Kinetic energy (eV)
    beta_0 : float, optional
        tilt angle offset

    Returns
    -------
    alpha : float or np.ndarray
        theta_par angle (analyser slit angle, rad)
    beta : float or np.ndarray
        tilt angle (rad)
    """
    # k_vacuum from KE
    kvac_str = "(KVAC_CONST * sqrt(Ek))"

    # Mapping function (include convert to degrees directly for speed)
    alpha = ne.evaluate(
        f"arcsin((sin(xi) * sqrt({kvac_str}**2 - ((kx * sin(delta_) - ky * cos(delta_))**2)) "
        f"- cos(xi) * (kx * sin(delta_) - ky * cos(delta_))) / {kvac_str})"
    )
    beta = ne.evaluate(
        f"(beta_0 + arctan((kx * cos(delta_) + ky * sin(delta_)) / sqrt({kvac_str}**2 - kx**2 - ky**2)))"
    )

    return alpha, beta


def _tij(ij):
    """
    Defines the inverse of the rotation matrix T_rot to obtain the elements of the inverse functions of type I' and II'
     manipulators, Eqn A9 of Ishida and Shin, Rev. Sci. Instrum. 89 (2018) 043903.

    Parameters
    ----------
    ij : int
        Index to return.

    Returns
    -------
    str
        String representation of relevant element of T_rot^-1 for passing to ne.evaluate().
    """

    expressions = {
        11: "cos(xi_) * cos(delta_)",
        12: "cos(xi_) * sin(delta_)",
        13: "-sin(xi_)",
        21: "(sin(chi_) * sin(xi_) * cos(delta_)) - (cos(chi_) * sin(delta_))",
        22: "(sin(chi_) * sin(xi_) * sin(delta_)) + (cos(chi_) * cos(delta_))",
        23: "sin(chi_) * cos(xi_)",
        31: "(cos(chi_) * sin(xi_) * cos(delta_)) + (sin(chi_) * sin(delta_))",
        32: "(cos(chi_) * sin(xi_) * sin(delta_)) - (sin(chi_) * cos(delta_))",
        33: "cos(chi_) * cos(xi_)",
    }

    return expressions[ij]


def _fIp_inv(kx, ky, delta_, xi_, chi_, Ek):  # Type I, with deflector
    """
    Convert k-space to angles for a type-I analyser with deflector, following the conventions and nomenclature of Ishida and Shin, Rev. Sci. Instrum. 89 (2018) 043903.

    Parameters
    ----------
    kx : float or np.ndarray
        k-vector along analyser slit (1/A)
    ky : float or np.ndarray
        k-vector perp to analyser slit (1/A)
    delta_ : float
        azi angle (rad), relative to 'normal' emission
    xi_ : float
        tilt angle (rad), relative to normal emission
    chi_ : float
        polar angle (rad), relative to normal emission
    Ek : float
        Kinetic energy (eV)

    Returns
    -------
    alpha : float or np.ndarray
        theta_par angle (analyser slit angle, rad)
    beta : float or np.ndarray
        polar angle (rad)
    """

    # k_vacuum from KE
    kvac_str = "(KVAC_CONST * sqrt(Ek))"
    k2p_str = f"sqrt({kvac_str}**2 - kx**2 - ky**2)"

    # Mapping functions
    arg1_str = (
        f"(({_tij(31)}) * kx) + (({_tij(32)}) * ky) + (({_tij(33)}) * ({k2p_str}))"
    )
    arg2_str = (
        f"(({_tij(11)}) * kx) + (({_tij(12)}) * ky) + (({_tij(13)}) * ({k2p_str}))"
    )
    arg3_str = (
        f"(({_tij(21)}) * kx) + (({_tij(22)}) * ky) + (({_tij(23)}) * ({k2p_str}))"
    )

    alpha = ne.evaluate(
        f"-arccos(({arg1_str}) / ({kvac_str})) * ({arg2_str}) / sqrt(({kvac_str})**2 - ({arg1_str})**2)"
    )
    beta = ne.evaluate(
        f"-arccos(({arg1_str}) / ({kvac_str})) * ({arg3_str}) / sqrt(({kvac_str})**2 - ({arg1_str})**2)"
    )

    return alpha, beta


def _fIIp_inv(kx, ky, delta_, xi_, chi_, Ek):  # Type II, with deflector
    """
    Convert k-space to angles for a type-II analyser with deflector, following the conventions and nomenclature of Ishida and Shin, Rev. Sci. Instrum. 89 (2018) 043903.

    Parameters
    ----------
    kx : float or np.ndarray
        k-vector perp to analyser slit (1/A)
    ky : float or np.ndarray
        k-vector along analyser slit (1/A)
    delta_ : float
        azi angle (rad), relative to 'normal' emission
    xi_ : float
        tilt angle (rad), relative to normal emission
    chi_ : float
        polar angle (rad), relative to normal emission
    Ek : float
        Kinetic energy (eV)

    Returns
    -------
    alpha : float or np.ndarray
        theta_par angle (analyser slit angle, rad)
    beta : float or np.ndarray
        mapping angle (rad)
    """

    # k_vacuum from KE
    kvac_str = "(KVAC_CONST * sqrt(Ek))"
    k2p_str = f"sqrt({kvac_str}**2 - kx**2 - ky**2)"

    # Mapping functions
    arg1_str = (
        f"(({_tij(31)}) * kx) + (({_tij(32)}) * ky) + (({_tij(33)}) * ({k2p_str}))"
    )
    arg2_str = (
        f"(({_tij(21)}) * kx) + (({_tij(22)}) * ky) + (({_tij(23)}) * ({k2p_str}))"
    )
    arg3_str = (
        f"(({_tij(11)}) * kx) + (({_tij(12)}) * ky) + (({_tij(13)}) * ({k2p_str}))"
    )

    alpha = ne.evaluate(
        f"arccos(({arg1_str}) / ({kvac_str})) * ({arg2_str}) / sqrt(({kvac_str})**2 - ({arg1_str})**2) "
    )
    beta = ne.evaluate(
        f"-arccos(({arg1_str}) / ({kvac_str})) * ({arg3_str}) / sqrt(({kvac_str})**2 - ({arg1_str})**2) "
    )

    return alpha, beta


def _f_kz(Ek, k_along_slit, k_perp_slit, V0):
    """Forward transform to kz.

    Parameters
    ----------
    Ek : float or np.ndarray
        Kinetic energy (eV).
    k_along_slit : float or np.ndarray
        k-vector along analyser slit (1/A).
    k_perp_slit : float or np.ndarray
        k-vector perp to analyser slit (1/A).
    V0 : float
        Inner potential (eV).

    Returns
    -------
    kz : float or np.ndarray
        k-vector for out-of-plan direction (1/A).
    """

    return ne.evaluate(
        "KVAC_CONST * sqrt(Ek - V0 - ((k_along_slit**2 + k_perp_slit**2)/(KVAC_CONST**2)))"
    )


def _f_inv_kz(kz, k_along_slit, k_perp_slit, V0, BE, wf):
    """Inverse transform from kz.

    Parameters
    ----------
    kz : float or np.ndarray
        k-vector for out-of-plan direction (1/A).
    k_along_slit : float or np.ndarray
        k-vector along analyser slit (1/A).
    k_perp_slit : float or np.ndarray
        k-vector perp to analyser slit (1/A).
    V0 : float
        Inner potential (eV).
    BE : float or np.ndarray
        Binding energy (eV).
    wf : float
        Work function (eV).

    Returns
    -------
    hv : float or np.ndarray
        Photon energy (eV).
    """

    return ne.evaluate(
        "(kz**2 + k_along_slit**2 + k_perp_slit**2) / (KVAC_CONST**2) + V0 + wf - BE"
    )


# --------------------------------------------------------- #
#      Helper functions for k-space conversion              #
# --------------------------------------------------------- #
def _reshape_for_2d(arr1, arr2):
    """
    Reshape arrays to make them broadcastable for 2D interpolation methods.

    Parameters
    ----------
    arr1 : np.ndarray
        First array to reshape.
    arr2 : np.ndarray
        Second array to reshape.

    Returns
    -------
    np.ndarray
        Reshaped array 1.
    np.ndarray
        Reshaped array 2.
    """

    arr1 = np.asarray(arr1).reshape(-1, 1)
    arr2 = np.asarray(arr2).reshape(1, -1)

    return arr1, arr2


def _reshape_for_3d(arr1, arr2, arr3):
    """
    Reshape arrays to make them broadcastable for 3D interpolation methods.

    Parameters
    ----------
    arr1 : np.ndarray
        First array to reshape.
    arr2 : np.ndarray
        Second array to reshape.
    arr3 : np.ndarray
        Third array to reshape.

    Returns
    -------
    np.ndarray
        Reshaped array 1.
    np.ndarray
        Reshaped array 2.
    np.ndarray
        Reshaped array 3.
    """

    arr1 = np.asarray(arr1).reshape(-1, 1, 1)
    arr2 = np.asarray(arr2).reshape(1, -1, 1)
    arr3 = np.asarray(arr3).reshape(1, 1, -1)

    return arr1, arr2, arr3


def _get_k_along_slit(kx, ky, ana_type):
    if ana_type in ["I", "Ip"]:
        return kx
    elif ana_type in ["II", "IIp"]:
        return ky
    else:
        raise ValueError(f"Invalid analyser type: {ana_type}")


def _get_k_perpto_slit(kx, ky, ana_type):
    if ana_type in ["I", "Ip"]:
        return ky
    elif ana_type in ["II", "IIp"]:
        return kx
    else:
        raise ValueError(f"Invalid analyser type: {ana_type}")


def get_kpar_cut(
    hv=21.2,
    Eb=0,
    theta_par_range=(-15, 15),
    polar=0,
    tilt=0,
    defl_perp=0,
    ana_type="I",
):
    """Extract the k_par cut along the analyser slit for a given set of angle parameters.

    Parameters
    ----------
    hv : float, optional
        Photon energy (eV) (default=21.2 eV).
    Eb : float, optional
        Binding energy (eV) (positive for below the Fermi level, default=0).
    theta_par_range : tuple, optional
        Range of theta_par values to calculate over in the form (start, stop);
        default=(-15, 15)).
    polar : float, optional
        Polar angle (deg) (default=0).
    tilt : float, optional
        Tilt angle (deg) (default=0).
    defl_perp : float, optional
        Deflector angle perpendicular to the slit (deg) (default=0).
    ana_type : str, optional
        Analyser type, one of I, II, Ip, IIp (default='I').
    """
    # Force deflector type if non-zero deflector angle
    if defl_perp and not ana_type.endswith("p"):
        ana_type = ana_type + "p"

    alpha = np.radians(np.linspace(*theta_par_range, 101))
    if ana_type == "I":
        beta, xi = np.radians(polar), np.radians(tilt)
        chi = None
    elif ana_type == "II":
        beta, xi = np.radians(tilt), np.radians(polar)
        chi = None
    else:
        beta, xi, chi = np.radians(defl_perp), np.radians(tilt), np.radians(polar)

    # Calculate kpar
    Ek = hv - 4.35 - Eb
    kx, ky = _f_dispatcher(ana_type, Ek, alpha, beta, 0, chi, 0, 0, 0, xi, 0)

    return kx, ky


def get_kz_cut(
    hv=21.2,
    Eb=0,
    polar_or_tilt=0,
    theta_par_range=(-15, 15),
    V0=12,
):
    """Extract the kz-dep of the cut along the analyser slit for given parameters.

    Parameters
    ----------
    hv : float, optional
        Photon energy (eV) (default=21.2 eV).
    Eb : float, optional
        Binding energy (eV) (positive for below the Fermi level, default=0).
    theta_par_range : tuple, optional
        Range of theta_par values to calculate over in the form (start, stop);
        default=(-15, 15)).
    polar_or_tilt : float, optional
        Angle offset in direction along the slit (deg) (default=0).
    V0 : float, optional
        Inner potential (eV) (default=12).
    """

    # Calculate for a type I analyser
    alpha = np.radians(np.linspace(*theta_par_range, 101))

    # Calculate kpar
    Ek = hv - 4.35 - Eb
    kx, ky = _f_dispatcher(
        "I", Ek, alpha, 0, 0, 0, 0, 0, 0, np.radians(polar_or_tilt), 0
    )
    kz = _f_kz(Ek, kx, ky, V0)

    return kx, kz


# --------------------------------------------------------- #
#      Main k-space conversion functions                    #
# --------------------------------------------------------- #


def k_convert(
    da,
    eV=None,
    eV_slice=None,
    kx=None,
    ky=None,
    kz=None,
    return_kz_scan_in_hv=False,
    quiet=False,
):
    """Perform k-conversion of angle dispersion or mapping data.

    Parameters
    ----------
    da : xarray.DataArray
        Data to convert to k-space.
    eV : slice, optional
        Binding energy range to calculate over for the final converted data in the form slice(start, stop, step).
        If not provided, the full energy range of the data will be used.
        Only energies within the limits of the data will be considered.
    eV_slice : float or tuple, optional
        Optional argument to return a single slice in energy, optionally integrated over some range, akin to the MDC
        method. Should be specified as a tuple of (energy, width) where energy is the central energy to integrate over
        and width is the total width of the integration window. Takes precedence over eV if provided.
    kx : slice, optional
        kx range to calculate over for the final converted data in the form slice(start, stop, step).
        Use `None` for any componet you do not wish to restrict, e.g. `(None,None,0.01)` to set the step size to 0.01.
        Only considered if kx is a dispersive direction of the data.
        Only kx values within the limits of the data will be considered.
        If not provided, the full kx range of the data will be used.
    ky : slice, optional
        ky range to calculate over for the final converted data in the form slice(start, stop, step).
        See Also: kx.
    kz : slice, optional
        kz range to calculate over for the final converted data in the form slice(start, stop, step).
        See Also: kx.
    return_kz_scan_in_hv : bool, optional
        If True, returns the converted data as (hv, eV, k_||) not (kz, eV, k_||).
    quiet : bool, optional
        If True, suppresses warnings and hides progress bar after k-space conversion completion.

    Returns
    -------
    xarray.DataArray
        Data converted to k-space.
    """

    # Parse basic data properties
    loader = BaseDataLoader.get_loader(da.metadata.scan.loc)
    angles = loader._get_angles_Ishida_Shin(da, quiet=quiet)  # Get angles for k-conv
    # Ensure all angles are in radians, da energies are in eV and then dequantify
    unit_stripped_angles = {
        axis: (angle.to("rad").magnitude if isinstance(angle, pint.Quantity) else angle)
        for axis, angle in angles.items()
    }
    angles = unit_stripped_angles
    da = da.pint.to({"hv": "eV", "eV": "eV"}).pint.dequantify()

    # Make a progressbar
    pb_steps = 3
    if "hv" in da.dims and not return_kz_scan_in_hv:
        pb_steps += 1
    pbar = tqdm(
        total=pb_steps,
        desc="Converting data to k-space - initialising",
        leave=not quiet,
    )

    # Get relevant energy scale information
    wf = _get_wf(da)  # Work fn, array if hv scan else single value
    if "hv" in da.dims:
        hv = da.hv.data
    else:
        hv = da.metadata.photon.hv.to("eV").magnitude

    BE_scale = _get_BE_scale(da)  # Tuple of start, stop, step
    # Restrict to manual energy range if specified
    if eV is not None:
        BE_scale = (
            np.max([eV.start if eV.start is not None else float("-inf"), BE_scale[0]]),
            np.min([eV.stop if eV.stop is not None else float("inf"), BE_scale[1]]),
            eV.step if eV.step is not None else BE_scale[2],
        )
    if eV_slice is not None:  # Get the full slice
        BE_scale = (
            eV_slice[0] - eV_slice[1] / 2,
            eV_slice[0] + eV_slice[1] / 2,
            BE_scale[2],
        )

    # Get bounds of data for k-conversion - use highest hv for hv scan
    if "hv" in da.dims and hv[-1] > hv[0]:
        EK_range = [hv[-1] + BE_scale[0] - wf[-1], hv[-1] + BE_scale[1] - wf[-1]]
    elif "hv" in da.dims and hv[-1] < hv[0]:
        EK_range = [hv[0] + BE_scale[0] - wf[0], hv[0] + BE_scale[1] - wf[0]]
    else:
        EK_range = [hv + BE_scale[0] - wf, hv + BE_scale[1] - wf]
    alpha_range = np.asarray(
        [
            np.min(angles["alpha"]),
            angles["xi_0"],
            np.max(angles["alpha"]),
        ]
    )

    # Check if a 2D or 3D conversion is required & reshape arrays to make them broadcastable
    if np.min(angles["beta"]) != np.max(angles["beta"]):
        beta_range = np.asarray([np.min(angles["beta"]), np.max(angles["beta"])])
        n_interpolation_dims = 3
        EK_range, alpha_range, beta_range = _reshape_for_3d(
            EK_range, alpha_range, beta_range
        )
    else:
        n_interpolation_dims = 2
        EK_range, alpha_range = _reshape_for_2d(EK_range, alpha_range)
        beta_range = angles["beta"]

    # Get k-space values corresponding to extremes of range
    kx_, ky_ = _f_dispatcher(
        ana_type=angles["type"],
        Ek=EK_range,
        alpha=alpha_range,
        beta=beta_range,
        beta_0=angles["beta_0"],
        chi=angles["chi"],
        chi_0=angles["chi_0"],
        delta=angles["delta"],
        delta_0=angles["delta_0"],
        xi=angles["xi"],
        xi_0=angles["xi_0"],
    )

    # Determine ranges
    k_along_slit = _get_k_along_slit(kx_, ky_, angles["type"])
    default_k_step = np.ptp(k_along_slit) / (len(angles["alpha"]) - 1)
    kx_range = (np.min(kx_), np.max(kx_) + default_k_step, default_k_step)
    ky_range = (np.min(ky_), np.max(ky_) + default_k_step, default_k_step)

    # Restrict to manual k ranges if specified and if a relevant axis
    if kx is not None and (n_interpolation_dims == 3 or angles["type"] in ["I", "Ip"]):
        kx_range = (
            np.max([kx.start if kx.start is not None else float("-inf"), kx_range[0]]),
            np.min([kx.stop if kx.stop is not None else float("inf"), kx_range[1]]),
            kx.step if kx.step is not None else kx_range[2],
        )
    if ky is not None and (n_interpolation_dims == 3 or angles["type"] in ["II", "IIp"]):
        ky_range = (
            np.max([ky.start if ky.start is not None else float("-inf"), ky_range[0]]),
            np.min([ky.stop if ky.stop is not None else float("inf"), ky_range[1]]),
            ky.step if ky.step is not None else ky_range[2],
        )

    # Make the arrays of required angle and energy values
    kx_values = np.arange(*kx_range)
    ky_values = np.arange(*ky_range)

    if "hv" not in da.dims:
        KE_values_no_curv = np.arange(*BE_scale) + hv - wf
    else:
        # For an hv scan, need to extract different KE values for each hv value,
        # but then flatten them for interpolation as we still only need bilinear interpolation
        KE_values_no_curv = (
            np.arange(*BE_scale).reshape(1, -1) + hv.reshape(-1, 1) - wf.reshape(-1, 1)
        )
        KE_values_no_curv_shape = KE_values_no_curv.shape  # Keep for later
        KE_values_no_curv = KE_values_no_curv.flatten()

    # Create meshgrid of angle and energy values for the interpolation
    if n_interpolation_dims == 3:
        KE_values_no_curv, kx_values, ky_values = _reshape_for_3d(
            KE_values_no_curv, kx_values, ky_values
        )
    else:
        if angles["type"] in ["II", "IIp"]:
            # Take the average k value of the perp to slit direction
            kx_values = np.mean(kx_values)
            KE_values_no_curv, ky_values = _reshape_for_2d(KE_values_no_curv, ky_values)
        else:
            KE_values_no_curv, kx_values = _reshape_for_2d(KE_values_no_curv, kx_values)
            ky_values = np.mean(ky_values)

    pbar.update(1)
    pbar.set_description_str(
        "Converting data to k-space - calculating inverse angle transformations"
    )

    alpha, beta = _f_inv_dispatcher(
        ana_type=angles["type"],
        Ek=KE_values_no_curv,
        kx=kx_values,
        ky=ky_values,
        beta_0=angles["beta_0"],
        chi=angles["chi"],
        chi_0=angles["chi_0"],
        delta=angles["delta"],
        delta_0=angles["delta_0"],
        xi=angles["xi"],
        xi_0=angles["xi_0"],
    )

    # Determine the KE values including curvature correction
    if n_interpolation_dims == 2:
        Ek_new = _get_E_shift_at_theta_par(
            da,
            alpha * 180 / np.pi,
            np.broadcast_to(
                KE_values_no_curv.reshape(-1, 1),
                (
                    KE_values_no_curv.size,
                    _get_k_along_slit(kx_values, ky_values, angles["type"]).size,
                ),
            ),
        )
    elif n_interpolation_dims == 3:
        Ek_new = _get_E_shift_at_theta_par(
            da,
            alpha * 180 / np.pi,
            np.broadcast_to(
                KE_values_no_curv.reshape(-1, 1, 1),
                (KE_values_no_curv.size, kx_values.size, ky_values.size),
            ),
        )

    # Interpolate onto the desired range
    pbar.update(1)
    pbar.set_description_str("Converting data to k-space - interpolating")
    # Check if we can use the faster rectilinear methods
    is_rectilinear = True
    for i in ["eV", "theta_par"]:
        if not _is_linearly_spaced(
            da[i].data, tol=(da[i].data[1] - da[i].data[0]) * 1e-3
        ):
            is_rectilinear = False

    if n_interpolation_dims == 2:
        if is_rectilinear:
            interpolation_fn = _fast_bilinear_interpolate_rectilinear
        else:
            interpolation_fn = _fast_bilinear_interpolate

        k_along_slit_label = _get_k_along_slit("kx", "ky", angles["type"])

        # Deal with any sign changes required for theta_par scale conventions and any stacked axes
        loader._get_sign_convention("theta_par")

        if "hv" in da.dims:
            interpolated_data = []
            for i, hv in enumerate(da.hv.data):
                data_hv = da.disp_from_hv(hv)
                start_index = i * KE_values_no_curv_shape[1]
                end_index = start_index + KE_values_no_curv_shape[1]
                interpolated_data_hv_slice = xr.apply_ufunc(
                    interpolation_fn,
                    Ek_new[start_index:end_index, :],
                    alpha[start_index:end_index, :],
                    data_hv.eV.data,
                    angles["alpha"],
                    data_hv,
                    input_core_dims=[
                        ["eV", k_along_slit_label],
                        ["eV", k_along_slit_label],
                        ["eV"],
                        ["theta_par"],
                        ["eV", "theta_par"],
                    ],
                    output_core_dims=[["eV", k_along_slit_label]],
                    exclude_dims={"eV"},
                    vectorize=True,
                    dask="parallelized",
                    keep_attrs=True,
                )
                interpolated_data.append(interpolated_data_hv_slice)
            interpolated_data = xr.concat(interpolated_data, dim="hv")
            # Add co-ordinates
            interpolated_data.coords.update(
                {
                    k_along_slit_label: _get_k_along_slit(
                        kx_values, ky_values, angles["type"]
                    ).squeeze(),
                    "eV": np.arange(*BE_scale),
                    "hv": da.hv.data,
                }
            )
            interpolated_data = interpolated_data.pint.quantify(
                {k_along_slit_label: "1/angstrom", "eV": "eV", "hv": "eV"}
            )

        else:
            interpolated_data = xr.apply_ufunc(
                interpolation_fn,
                Ek_new,
                alpha,
                da.eV.data,
                angles["alpha"],
                da,
                input_core_dims=[
                    ["eV", k_along_slit_label],
                    ["eV", k_along_slit_label],
                    ["eV"],
                    ["theta_par"],
                    ["eV", "theta_par"],
                ],
                output_core_dims=[["eV", k_along_slit_label]],
                exclude_dims={"eV"},
                vectorize=True,
                dask="parallelized",
                keep_attrs=True,
            )
            # Add co-ordinates
            interpolated_data.coords.update(
                {
                    k_along_slit_label: _get_k_along_slit(
                        kx_values, ky_values, angles["type"]
                    ).squeeze(),
                    "eV": np.arange(*BE_scale),
                }
            )
            interpolated_data = interpolated_data.pint.quantify(
                {k_along_slit_label: "1/angstrom", "eV": "eV"}
            )
        # Add the perpendicular momentum to the data attributes
        interpolated_data.attrs[_get_k_perpto_slit("kx", "ky", angles["type"])] = (
            _get_k_perpto_slit(kx_values, ky_values, angles["type"]).squeeze()
        )
    else:  # Should be Fermi map
        # Other angular dimension - assume the only remaining dimension
        other_dim = list(set(da.dims) - set(["eV", "theta_par"]))[0]

        # Check linearity of remaining dimension
        is_rectilinear = is_rectilinear and _is_linearly_spaced(
            da[other_dim].data,
            tol=(da[other_dim].data[1] - da[other_dim].data[0]) * 1e-3,
        )

        if is_rectilinear:
            interpolation_fn = _fast_trilinear_interpolate_rectilinear
        else:
            interpolation_fn = _fast_trilinear_interpolate

        with numba_progress.ProgressBar(
            total=Ek_new.size,
            dynamic_ncols=True,
            delay=0.2,
            desc="Interpolating onto new grid",
            leave=False,
        ) as nb_pbar:
            interpolated_data = xr.apply_ufunc(
                interpolation_fn,
                Ek_new,
                alpha,
                beta,
                da.eV.data,
                angles["alpha"],
                angles["beta"],
                da,
                nb_pbar,
                input_core_dims=[
                    ["eV", "kx", "ky"],
                    ["eV", "kx", "ky"],
                    ["eV", "kx", "ky"],
                    ["eV"],
                    ["theta_par"],
                    [other_dim],
                    ["eV", "theta_par", other_dim],
                    [],
                ],
                output_core_dims=[["eV", "kx", "ky"]],
                exclude_dims={"eV"},
                vectorize=True,
                dask="parallelized",
                keep_attrs=True,
            ).transpose(
                _get_k_perpto_slit("kx", "ky", angles["type"]),
                "eV",
                _get_k_along_slit("kx", "ky", angles["type"]),
            )
        interpolated_data.coords.update(
            {
                "kx": kx_values.squeeze(),
                "eV": np.arange(*BE_scale),
                "ky": ky_values.squeeze(),
            }
        )
        interpolated_data = interpolated_data.pint.quantify(
            {"kx": "1/angstrom", "eV": "eV", "ky": "1/angstrom"}
        )

    if "hv" in da.dims and not return_kz_scan_in_hv:
        # If hv scan and if required, do kz-interpolation step to return the data in the form (kz, eV, k_||)
        pbar.update(1)
        pbar.set_description_str("Converting data to k-space - converting to kz")
        # Convert to kz
        interpolated_data = _convert_to_kz(
            interpolated_data, kz, wf, [np.min(Ek_new), np.max(Ek_new)]
        )

    # Do a hack to remove some noise at the boundary which can give negative values, screwing up the plots
    if da.min() <= 0:
        interpolated_data = interpolated_data.where(interpolated_data > 0, 0)

    # Update the energy type in data attributes
    interpolated_data.metadata.analyser.scan.eV_type = "Binding Energy"

    # Update the history
    hist_str = "Converted to k-space using the following parameters: "
    reference_angles = {k: v for k, v in angles.items() if ("_0" in k and v is not None)}
    hist_str += f"Reference angles: {reference_angles}, "
    if "hv" in da.dims:
        hist_str += f"Inner potential: {da.metadata.calibration.V0 or 12 * ureg('eV')}, "
    hist_str += f"Time taken: {pbar.format_dict['elapsed']:.2f}s."
    interpolated_data.history.add(hist_str)

    if eV_slice is not None:
        interpolated_data = interpolated_data.mean("eV")
        interpolated_data.history.add(
            f"Data integrated in energy about {eV_slice[0]} eV +/- {eV_slice[1] / 2} eV."
        )

    pbar.update(1)
    pbar.set_description_str("Converting data to k-space - complete")

    if pbar.format_dict["elapsed"] < 0.2:  # Hide if only ran for a short time
        pbar.leave = False
    pbar.close()

    return interpolated_data.pint.quantify()


def _convert_to_kz(da, kz, wf, Ek_range):
    # Get relevant parameters
    k_along_slit_str = list(set(da.dims) - {"hv", "eV"})[0]
    k_perp_slit_str = "ky" if k_along_slit_str == "kx" else "kx"
    k_perp_slit = da.attrs.get(k_perp_slit_str, 0)
    k_along_slit = da[k_along_slit_str].data
    V0 = da.metadata.calibration.V0
    if V0 is None:
        V0 = 12 * ureg("eV")
        analysis_warning(
            f"No inner potential provided, defaulting to V0={V0}.",
            "warning",
            "Missing inner potential",
        )
    V0 = V0.to("eV").magnitude

    # Get kz values corresponding to the extremes of the range
    kz_ = _f_kz(
        np.asarray(Ek_range).reshape(-1, 1),
        np.asarray([min(k_along_slit), 0, max(k_along_slit)]).reshape(1, -1),
        k_perp_slit,
        V0,
    )
    # Determine the kz values to interpolate onto, including manual range if specified
    default_k_step = (np.nanmax(kz_) - np.nanmin(kz_)) / (
        2 * len(da.hv)
    )  # Default is based on 0.5 step from the number of hv points
    kz_range = (np.nanmin(kz_), np.nanmax(kz_), default_k_step)
    # Restrict to manual kz range if specified
    if kz is not None:
        kz_range = (
            np.max([kz.start if kz.start is not None else float("-inf"), kz_range[0]]),
            np.min([kz.stop if kz.stop is not None else float("inf"), kz_range[1]]),
            kz.step if kz.step is not None else kz_range[2],
        )
    kz_values = np.arange(kz_range[0], kz_range[1] + kz_range[2], kz_range[2])

    # Do inverse transform to get the required hv values for interpolation [kz, eV, k_||]
    # Take the first work function value for getting this hv scale - need to account for this in the final interpolation
    hv_values = _f_inv_kz(
        kz_values.reshape(-1, 1, 1),
        k_along_slit.reshape(1, 1, -1),
        k_perp_slit,
        V0,
        da.eV.data.reshape(1, -1, 1),
        wf[0],
    )

    # Get true hv values accounting for the change which was previously encoded in an effective wf change
    wf_diff = wf - wf[0]
    true_hv = da.hv.data + wf_diff

    # Do the interpolation, data originally [hv, eV, k_||]
    k_along_slit_vectorised = np.broadcast_to(
        k_along_slit.reshape(1, 1, -1), hv_values.shape
    )
    BE_vectorised = np.broadcast_to(da.eV.data.reshape(1, -1, 1), hv_values.shape)
    with numba_progress.ProgressBar(
        total=hv_values.size,
        dynamic_ncols=True,
        delay=0.2,
        desc="Interpolating onto kz grid",
        leave=False,
    ) as nb_pbar:
        interpolated_data = xr.apply_ufunc(
            _fast_trilinear_interpolate,
            hv_values,
            BE_vectorised,
            k_along_slit_vectorised,
            true_hv,
            da.eV.data,
            k_along_slit,
            da.pint.dequantify(),
            nb_pbar,
            input_core_dims=[
                ["kz", "eV", k_along_slit_str],
                ["kz", "eV", k_along_slit_str],
                ["kz", "eV", k_along_slit_str],
                ["hv"],
                ["eV"],
                [k_along_slit_str],
                ["hv", "eV", k_along_slit_str],
                [],
            ],
            output_core_dims=[["kz", "eV", k_along_slit_str]],
            exclude_dims={},
            vectorize=True,
            dask="parallelized",
            keep_attrs=True,
        )

        # Add the kz axis to the data
        interpolated_data.coords.update(
            {
                "kz": kz_values.squeeze(),
            }
        )

    return interpolated_data.pint.quantify({"kz": "1/angstrom"})



"""Static in-line plotting functions."""

import matplotlib.pyplot as plt
import numpy as np
import panel as pn
import pint_xarray  # noqa: F401
import xarray as xr
from cycler import cycler
from IPython.display import display
from matplotlib import cm, colors
from numpy.fft import fft

from peaks.core.utils.misc import analysis_warning


def plot_grid(
    data,
    ncols=3,
    nrows=None,
    titles=None,
    sharex=False,
    sharey=False,
    figsize=None,
    vmin=None,
    vmax=None,
    cmap=None,
    **plotting_kwargs,
):
    """Plots an array of 2D DataArrays on a grid.

    Parameters
    ------------
    data : list or xarray.DataTree
         A list or :class:`xarray.DataTree` containing 2D scans (in :class:`xarray.DataArray` format) to be plotted.
         If a datatree, the tree should be hollow, and the leaves should be :class:`xarray.Dataset`'s with a single
         data variable, `data`

    ncols : int, optional
        Number of columns. Ignored if nrows is specified. Defaults to 3 (or lower if <3 plots).

    nrows : int, optional
        Number of rows. Overwrites ncols. Defaults to None (i.e. ncols is used).

    titles : list, optional
        List of subtitles to be supplied on the plots. Length must match length of the data list. Defaults to None.

    sharex : bool, optional
        Whether to have the plots share an x-axis. Defaults to False.

    sharey : bool, optional
        Whether to have the plots share a y-axis. Defaults to False.

    figsize : tuple, optional
        Size of figure to be plotted. Defaults to None.

    vmin : float or int or list, optional
        Minimum value for the colorbar. Use a list to specify separate values for each plot. If a list is passed, the
        length must match the number of plots. Defaults to None.

    vmax : float or int or list, optional
        Maximum value for the colorbar. Use a list to specify separate values for each plot. If a list is passed, the
        length must match the number of plots. Defaults to None.

    cmap : str or list, optional
        Matplotlib colormap to use for the plots. Use a list to specify separate colormaps for each plot. If a list is
        passed, the length must match the number of plots.

    **plotting_kwargs
        Additional standard matplotlib calls arguments to pass to the plot.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        disp1 = load('disp1.ibw')
        disp2 = load('disp2.ibw')
        disp3 = load('disp3.ibw')
        disp4 = load('disp4.ibw')

        disps_to_plot = [disp1, disp2, disp3, disp4]
        disp_titles = ['10 K', '20 K', '30 K', '40 K']

        # Plot dispersions on a 2D grid
        plot_grid(disps_to_plot)

        # Plot dispersions on a 2D grid, where each dispersion has a title, all dispersions share an x-axis, and
        # colour bars are not shown
        plot_grid(disps_to_plot, titles=disp_titles, sharex=True, add_colorbar=False)

    """

    # Number of plots
    nplots = len(data)

    # Check list-like inputs are the correct length
    for item in [vmin, vmax, cmap]:
        if isinstance(item, (list, np.ndarray)):
            if len(item) != nplots:
                raise ValueError(
                    f"Length of supplied {item} list does not match number of plots. Either specify a list of length "
                    f"{nplots} or pass a single value."
                )

    # Check default columns is sensible
    if nplots < ncols:
        ncols = nplots

    # Number of rows required
    if nrows:  # If nrows specified in call
        ncols = int(np.ceil(nplots / nrows))  # Overwrite default columns call
    else:
        nrows = int(np.ceil(nplots / ncols))  # Set nrows based on ncols and nplots

    # If figsize not specified, set a default based on number of rows and columns
    if not figsize:
        figsize = (5 * ncols, 5 * nrows)

    # Make the figure layout
    fig, axes = plt.subplots(
        ncols=ncols, nrows=nrows, figsize=figsize, sharex=sharex, sharey=sharey
    )

    # Clear any remaining subplot axes
    if nrows * ncols > nplots:
        for i in range(nplots, nrows * ncols):
            # Work out grid positions
            j0 = int(np.floor((i) / ncols))
            j1 = int(i - ncols * j0)
            axes[j0][j1].axis("off")

    # Check for whether plot titles are to be displayed
    plot_titles = False
    if titles:
        if len(titles) == nplots:  # Check the correct number are provided
            plot_titles = True
        else:  # If not, don't use them and give a warning
            analysis_warning(
                "Length of supplied titles list does not match number of plots. Supplied titles have not been used.",
                title="Plotting info",
                warn_type="danger",
            )

    # Make the plots
    def _plot_single(da, count, vmin, vmax, cmap):
        additional_plotting_kwargs = {
            k: v
            for k, v in {"vmin": vmin, "vmax": vmax, "cmap": cmap}.items()
            if v is not None
        }
        if nrows < 2 or ncols == 1:  # 1D grid
            ax = axes[count]
        else:  # 2D grid
            j0, j1 = divmod(count, ncols)
            ax = axes[j0][j1]
        da.plot(ax=ax, **plotting_kwargs, **additional_plotting_kwargs)  # Plot data
        if plot_titles:  # If plot titles to be displayed, update them here
            ax.set_title(titles[count])

    def _get_value_from_list(value, count):
        return value[count] if isinstance(value, list) else value

    if isinstance(data, list):
        for count, value in enumerate(data):
            _plot_single(
                value,
                count,
                _get_value_from_list(vmin, count),
                _get_value_from_list(vmax, count),
                _get_value_from_list(cmap, count),
            )
    elif isinstance(data, xr.DataTree):
        # If titles not passed as a specific list, we will populate them from node names
        if not plot_titles:
            titles = []
            plot_titles = True
        offset = 0
        for count, node in enumerate(data.subtree):
            if not node.is_empty:
                titles.append(node.name)
                _plot_single(
                    node.data,
                    count - offset,
                    _get_value_from_list(vmin, count - offset),
                    _get_value_from_list(vmax, count - offset),
                    _get_value_from_list(cmap, count - offset),
                )
            else:
                offset += 1

    # Tidy up the layout
    plt.tight_layout()


def plot_DCs(
    DCs,
    titles=None,
    cmap="coolwarm",
    color=None,
    offset=0,
    norm=False,
    stack_dim="auto",
    figsize=(6, 6),
    linewidth=1,
    ax=None,
    **plotting_kwargs,
):
    """Plot a DC stack with the colours varying according to a colormap.

    Parameters
    ------------
    DCs : xarray.DataArray, list
         DCs to plot, either as a list of single DC DataArrays, or a single DataArray with the stack of DCs included.

    titles : list, optional
        List of DC labels to be used in the legend. Length must match number of DCs to plot. Defaults to None.

    cmap : str, optional
        Matplotlib cmap to use for line colors. Defaults to coolwarm.

    color : str, optional
        Single matplotlib color to use for all lines. Takes precedence over cmap if defined. Defaults to None.

    offset : float, optional
        Vertical offsets between subsequent DCs, represented as a fraction of the maximum peak height. Defaults to 0.

    norm : bool, optional
        Whether to normalise the DCs to 1. Defaults to False.

    stack_dim : str, optional
        Which dimension to stack DCs along. Ignored if a list of DCs is passed. Defaults to 'auto' which takes the
        smallest dimension.

    figsize : tuple, optional
        Size of figure to be plotted. Defaults to None.

    linewidth : float, optional
        Width of lines to be plotted. Defaults to 1.

    ax : numpy.ndarray, optional
        Specific matplotlib axis call. Defaults to None.

    **plotting_kwargs
        Additional standard matplotlib calls arguments to pass to the plot.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        disp1 = pks.load('disp1.ibw')

        EDCs_to_plot = disp.EDC(k=[0, 0.1, 0.2, 0.3, 0.4, 0.5], dk=0.01)

        # Plot EDCs
        EDCs_to_plot.plot_DCs()

        MDC_1 = disp.MDC(E=95.61, dE=0.01)
        MDC_2 = disp.MDC(E=95.62, dE=0.01)
        MDC_3 = disp.MDC(E=95.63, dE=0.01)

        # Plot normalised MDCs with a 5% (of maximum height) offset, and add legend
        plot_DCs([MDC_1, MDC_2, MDC_3], titles = ['EF', 'EF - 0.01 eV', 'EF -0.02 eV'], norm=True, offset=0.05)

    """

    # If a list of DCs is supplied, combine them into a single DataArray
    if isinstance(DCs, list):
        DC_array = xr.concat(DCs, dim="DC_no")
        stack_dim = "DC_no"

    elif isinstance(DCs, xr.DataArray):
        # If a DataArray is instead supplied, finding the stacking dimension if not defined
        DC_array = DCs.copy(deep=True)  # Make a copy so we can safely modify
        if stack_dim == "auto":
            # Assume the stacking dimension is the smallest dimension size
            if DC_array.shape[0] < DC_array.shape[1]:
                stack_dim = DC_array.dims[0]
            else:
                stack_dim = DC_array.dims[1]

    elif isinstance(DCs, xr.DataTree):
        DC_list = [DC.data.copy(deep=True) for DC in DCs.subtree if not DC.is_empty]
        DC_array = xr.concat(DC_list, dim="DC_no")
        stack_dim = "DC_no"
    else:
        raise ValueError(
            "DCs must be a list of xarray.DataArray's, a single xarray.DataArray containing the DCs to "
            "plot, or a xarray.DataTree of DCs to plot."
        )

    DC_array = DC_array.pint.dequantify()

    # Check correct dimension supplied
    if len(DC_array.shape) != 2:
        raise Exception(
            "Incorrect dimension of data supplied. Please supply a single 2D DataArray or list of 1D DataArrays."
        )

    # Get non-stacking dimension
    other_dim = [dim for dim in DC_array.dims if dim != stack_dim][0]

    # Normalise if required
    if norm:
        for coord in DC_array[stack_dim]:
            DC_array.loc[{stack_dim: coord}] = (
                DC_array.loc[{stack_dim: coord}] / DC_array.loc[{stack_dim: coord}].max()
            )

    # Get absolute offset from fractional offset
    absolute_offset = offset * float(DC_array.max())

    # Define linear offset wave in steps of the supplied or guessed offset
    offset_data = xr.DataArray(
        data=np.arange(len(DC_array[stack_dim])) * absolute_offset,
        dims=stack_dim,
        coords={stack_dim: DC_array[stack_dim]},
    )

    # Add offset to DC data
    DC_array += offset_data

    # If no tiles are provided, no legend will be displayed
    if not titles:
        plotting_kwargs["add_legend"] = False

    # If x and y are not defined, make a guess for which axes to plot the dimensions on.
    if "y" not in plotting_kwargs and "x" not in plotting_kwargs:
        # If it seems to be an EDC plot, set vertical by default
        if other_dim == "eV":
            plotting_kwargs["y"] = "eV"
        # For anything else, set horizontal
        else:
            plotting_kwargs["x"] = other_dim

    # Set up plot
    if ax:
        plotting_kwargs["ax"] = ax
    else:
        plt.figure(figsize=figsize)
        ax = plt.gca()

    # If the user has requested to plot lines of a single color
    if color:
        DC_array.plot.line(
            hue=stack_dim, linewidth=linewidth, color=color, **plotting_kwargs
        )

    # Else use a colormap
    else:
        # Define the colour scheme
        cols = getattr(cm, cmap)(np.linspace(0, 1, len(DC_array[stack_dim])))
        cmap = colors.ListedColormap(cols)
        custom_cycler = cycler(color=cmap.colors)
        ax.set_prop_cycle(custom_cycler)
        DC_array.plot.line(hue=stack_dim, linewidth=linewidth, **plotting_kwargs)

    # Add axis titles
    if "y" in plotting_kwargs:
        plt.xlabel("Intensity [arb. units]")
    else:
        plt.ylabel("Intensity [arb. units]")

    # Add legend if user has inputted titles
    if titles:
        if len(titles) == len(DC_array[stack_dim]):
            for ct, i in enumerate(ax.lines):
                i.set_label(titles[ct])
            ax.legend()
        else:
            analysis_warning(
                "Length of supplied titles list does not match number of DCs. Supplied titles have not been used.",
                title="Plotting info",
                warn_type="danger",
            )

    # Tidy up the layout
    plt.tight_layout()


def plot_3d_stack(
    da,
    downsample=2,
    stack_dim="eV",
    vmax=None,
    cmap="Purples",
    figsize=(8, 12),
    aspect=None,
    elev=10.0,
    azim=-60.0,
    **kwargs,
):
    """Plot a stack of surface plots from a data cube

    Parameters
    ------------
    da : xarray.DataArray
        The data to plot.
    downsample : int, optional
        The downsample factor for the plot. Defaults to 2.
    stack_dim : str, optional
        The dimension to stack the data along. Defaults to 'eV'.
    vmax : float, optional
        The maximum value for the color scale. Defaults to None (i.e. the maximum value
        of the data).
    cmap : str, optional
        The colormap to use. Defaults to 'Purples'.
    figsize : tuple, optional
        The size of the figure. Defaults to (8, 12).
    aspect : tuple, optional
        The aspect ratio for the plot (x, y, z), in units of the data, e.g. (1,1,200)
        Defaults to None.
    elev : float, optional
        The elevation angle for the 3D plot. Defaults to 10.0.
    azim : float, optional
        The azimuthal angle for the 3D plot. Defaults to -60.0.
    **kwargs : optional
        Additional standard matplotlib calls arguments to pass to the plot.
        Must be suitable for passing in ax.set(**kwargs) format.

    Notes
    ------------
    `xlim` and `ylim` don't clip the range for this type of plot. Pass data over the
    range you want to plot. Similarly, apply any normalisation to the data before
    passing it to this function.
    """

    # Parse the dimensions
    if stack_dim not in da.dims:
        raise ValueError(
            f"stack_dim '{stack_dim}' is not a dimension of the data array."
        )
    non_stack_dims = [d for d in da.dims if d != stack_dim]

    # Normalise the data
    if not vmax:  # no max colour limit specified
        vmax = da.max().pint.dequantify().data
    C = (da / vmax).copy().pint.dequantify()
    C = np.clip(C, 0, 1)

    # Set up the plot
    plt.interactive(True)
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111, projection="3d")
    cmap_sel = plt.get_cmap(cmap)

    # Define the grid
    X = da[non_stack_dims[1]].data
    Y = da[non_stack_dims[0]].data
    XX, YY = np.meshgrid(X, Y)
    Z = np.zeros_like(XX)

    # Set the aspect ratio
    if aspect:
        stack_dim_range = np.ptp(da[stack_dim].data)
        ax.set_box_aspect(
            (np.ptp(X) * aspect[0], np.ptp(Y) * aspect[1], stack_dim_range * aspect[2])
        )

    # Make the plot surfaces (downsample 2x2)
    for i in da["eV"].data:
        ax.plot_surface(
            XX,
            YY,
            Z + i,
            rstride=downsample,
            cstride=downsample,
            facecolors=cmap_sel(C.sel(eV=i).data),
            shade=False,
        )

    # Colour scale
    m = cm.ScalarMappable(cmap=cmap)
    m.set_array(np.linspace(0, vmax, 100))
    cbar = plt.colorbar(m, ax=ax, shrink=0.2, aspect=8, pad=0.15, location="right")
    cbar.ax.yaxis.set_ticks_position("right")
    cbar.ax.yaxis.set_label_position("right")

    # Some plot display settings
    # Label the axes
    ax.set_xlabel(non_stack_dims[1])
    ax.set_ylabel(non_stack_dims[0])
    ax.set_zlabel(stack_dim)

    # Set overall view (no grids etc.)
    ax.grid(False)
    ax.xaxis.pane.set_edgecolor("black")
    ax.yaxis.pane.set_edgecolor("black")
    ax.xaxis.pane.fill = False
    ax.yaxis.pane.fill = False
    ax.zaxis.pane.fill = False

    # Sort out axis label locations
    [t.set_va("center") for t in ax.get_yticklabels()]
    [t.set_ha("left") for t in ax.get_yticklabels()]
    [t.set_va("center") for t in ax.get_xticklabels()]
    [t.set_ha("right") for t in ax.get_xticklabels()]
    [t.set_va("center") for t in ax.get_zticklabels()]
    [t.set_ha("left") for t in ax.get_zticklabels()]
    ax.xaxis.set_rotate_label(False)
    ax.yaxis.set_rotate_label(False)
    ax.zaxis.set_rotate_label(False)
    ax.zaxis.labelpad = 20

    # Set z-axis ticks to match the slices
    ax.zaxis.set_major_locator(plt.FixedLocator(np.round(C["eV"].data, 3)))

    # Set default elevation view
    ax.view_init(elev=elev, azim=azim)

    ax.set(**kwargs)

    # Show the plot
    plt.show()


def plot_fit(fit_results_ds, show_components=True, figsize=None, **kwargs):
    def _plot_single_fit(fit_results, show_components, figsize, **kwargs):
        fig = plt.figure(figsize=figsize)
        for dim in fit_results.dims:
            if dim in kwargs:
                fit_results = fit_results.isel({dim: kwargs.pop(dim)})
        # Add the independent variable to the kwargs for plotting
        kwargs.setdefault("xlabel", fit_results_ds.attrs.get("independent_var", "x"))

        fit_model = fit_results["fit_model"].compute().item()
        fit_model.plot(fig=fig, **kwargs)
        if show_components and len(fit_model.components) > 1:
            ax = plt.gca()
            components = fit_model.eval_components()
            for component_name, component_data in components.items():
                if component_name != "_gauss_conv":
                    ax.plot(
                        fit_model.userkws["x"],
                        component_data,
                        label=component_name,
                        linestyle="--",
                    )
            ax.legend()
        plt.close(fig)
        return fig

    # Check the data array contains fit results
    if "fit_model" not in fit_results_ds:
        raise ValueError(
            "The passed data does not appear to be a DataSet containing fit results. Generate the relevant fit results "
            "by calling the `fit` method on a suitable DataArray, e.g. "
            "`fit_results = disp1.fit(model, 'eV', params)`"
        )

    if len(fit_results_ds["fit_model"].shape) == 0:
        fig = _plot_single_fit(fit_results_ds, show_components, figsize, **kwargs)
        display(fig)
    else:
        # Initialize Panel extension
        pn.extension()

        # Create sliders for dims
        sliders = {}
        for dim in fit_results_ds.dims:
            sliders[dim] = pn.widgets.IntSlider(
                name=dim, start=0, end=len(fit_results_ds[dim]) - 1, step=1, value=0
            )

        # Bind the plot function to the sliders, updating the plot dynamically
        interactive_plot = pn.bind(
            _plot_single_fit,
            fit_results=fit_results_ds,
            show_components=show_components,
            figsize=figsize,
            **sliders,
            **kwargs,
        )

        # Display the sliders and the plot in the notebook
        dashboard = pn.Column(
            pn.Row(*sliders.values()), pn.pane.Matplotlib(interactive_plot)
        )

        dashboard.servable()
        return dashboard


def plot_fit_test(data, model, params, show_components=True, **kwargs):
    """Compare a fit model evaluated for some fit parameters to a 1D data array.

    Parameters
    ------------
    data : xarray.DataArray
        The data to compare the model to.
    model : lmfit.Model
        The model to evaluate.
    params : lmfit.Parameters
        The parameters to evaluate the model with.
    show_components : bool, optional
        Whether to show the individual components of the model. Defaults to True.
    **kwargs : optional
        Additional standard matplotlib calls arguments to pass to the plot.
    """

    data = data.pint.dequantify()

    if len(data.dims) != 1:
        raise ValueError(
            "Data must be 1D with the dimension corresponding to the indpependent variable."
        )

    # Evaluate the model
    model_result = model.eval(params=params, x=data[data.dims[0]].data)
    if show_components:
        components = model.eval_components(params=params, x=data[data.dims[0]].data)

    # Plot the data and the model
    data.plot(label="Data", marker="o", **kwargs)
    plt.plot(data[data.dims[0]].data, model_result, label="Model", **kwargs)
    if show_components:
        for component_name, component_data in components.items():
            if component_name != "_gauss_conv":
                plt.plot(
                    data[data.dims[0]].data,
                    component_data,
                    label=component_name,
                    linestyle="--",
                    **kwargs,
                )
    plt.legend()
    plt.show()


def plot_ROI(
    ROI,
    color="black",
    x=None,
    y=None,
    label=None,
    loc="best",
    ax=None,
    **plotting_kwargs,
):
    """This function plots a region of interest (ROI).

    Parameters
    ------------
    ROI : dict
        A dictionary of two lists which contains the vertices of the polygon for the ROI definition, in the form
        {'dim1': [pt1, pt2, pt3, ...], 'dim2'=[pt1', pt2', pt3', ...]}. As many points can be specified as required,
        but this should be given with the same number of points for each dimension.

    color : str, optional
        Matplotlib color to use for ROI plot. Defaults to 'black'.

    x : str, optional
        Specifies which coordinate to plot along the x-axis. Takes precedence over y (only one of x and y need to be
        specified, the other is set automatically). Defaults to the first coordinate listed in the ROI dictionary.

    y : str, optional
        Specifies which coordinate to plot along the y-axis. Defaults to the second coordinate listed in the ROI
        dictionary.

    label : str, optional
        ROI label to pass to a legend. Defaults to None (where no legend is plotted).

    loc : str, int, optional
        Standard matplotlib call to specify legend location. Defaults to 'best'.

    ax : numpy.ndarray, optional
        Specific matplotlib axis call. Defaults to None.

    **plotting_kwargs
        Additional standard matplotlib calls arguments to pass to the plot.

    Examples
    ------------
    Example usage is as follows::

    from peaks import *

        disp = load('disp1.ibw')

        ROI1 = {'theta_par': [-8, -5.5, -3.1, -5.6], 'eV': [95.45, 95.45, 95.77, 95.77]}
        ROI2 = {'theta_par': [-9, -4, -2.5, -5.9], 'eV': [95.45, 95.45, 95.77, 95.77]}

        # Plot ROI
        plot_ROI(ROI1)
        plt.show()

        # Plot ROI on top of a dispersion with eV plotted along the x-axis.
        disp1.plot()
        plot_ROI(ROI1, x='eV')
        plt.show()

        # Plot ROIs on top of dispersions on specific axes, axes[0] and axes[1], with labels
        fig, axes= plt.subplots(ncols=2)
        disp.T.plot(ax=axes[0])
        plot_ROI(ROI1, label='ROI 1', ax=axes[0])
        disp.T.plot(ax=axes[1])
        plot_ROI(ROI2, label='ROI 2', ax=axes[1])

    """

    # Determine relevant dimensions for ROI
    dims = list(ROI)

    # Determine lists of vertices of ROI
    verts0 = ROI[dims[0]]
    verts1 = ROI[dims[1]]

    # Append initial value to end to close polygon
    verts0.append(verts0[0])
    verts1.append(verts1[0])

    # By default, plot the second dimension as the y-axis
    yax = 1

    # Check if definition of x- or y-axis is given
    if x:
        yax = 1 - dims.index(x)
    elif y:
        yax = dims.index(y)

    # Plot the ROI
    if ax:  # If particular axis specified
        if yax == 0:
            ax.plot(verts1, verts0, color=color, label=label, **plotting_kwargs)
        else:
            ax.plot(verts0, verts1, color=color, label=label, **plotting_kwargs)
    else:  # Otherwise just call with plt
        if yax == 0:
            plt.plot(verts1, verts0, color=color, label=label, **plotting_kwargs)
        else:
            plt.plot(verts0, verts1, color=color, label=label, **plotting_kwargs)

    # If called with label, plot the legend
    if label:
        if ax:  # If particular axis specified
            ax.legend(loc=loc)
        else:  # Otherwise just call with plt
            plt.legend(loc=loc)

    # Tidy up the layout
    plt.tight_layout()


def plot_kpar_cut(
    hv=21.2,
    Eb=0,
    theta_par_range=(-15, 15),
    polar=0,
    tilt=0,
    defl_perp=0,
    ana_type="I",
    ax=None,
    label_cut=True,
    flip=False,
    **kwargs,
):
    """Plot the k_par cut along the analyser slit for a given set of angle parameters.

    Parameters
    ----------
    hv : float, optional
        Photon energy (eV) (default=21.2 eV).
    Eb : float, optional
        Binding energy (eV) (positive for below the Fermi level, default=0).
    theta_par_range : tuple, optional
        Range of theta_par values to calculate over in the form (start, stop);
        default=(-15, 15)).
    polar : float, optional
        Polar angle (deg) (default=0).
    tilt : float, optional
        Tilt angle (deg) (default=0).
    defl_perp : float, optional
        Deflector angle perpendicular to the slit (deg) (default=0).
    ana_type : str, optional
        Analyser type, one of I, II, Ip, IIp (default='I').
    ax : matplotlib.axes.Axes, optional
        Matplotlib axes object to plot on (default=None).
    label_cut : bool, optional
        Whether to label the cut with the angles used (default=True).
    flip : bool, optional
        Whether to flip the kx and ky axes (default=False).
    **kwargs
        Additional keyword arguments to pass to the plot.
    """

    from peaks.core.process.k_conversion import get_kpar_cut

    kx, ky = get_kpar_cut(
        hv=hv,
        Eb=Eb,
        theta_par_range=theta_par_range,
        polar=polar,
        tilt=tilt,
        defl_perp=defl_perp,
        ana_type=ana_type,
    )

    if not ax:
        fig, ax = plt.subplots()
    if not flip:
        ax.plot(kx, ky, **kwargs)
    else:
        ax.plot(ky, kx, **kwargs)

    if label_cut:
        # Add label to the line showing the tilt and polar angles
        angle_label = f"Pol.: {polar}°, Tilt: {tilt}°"
        label_pos = [kx[-1], ky[-1]] if not flip else [ky[-1], kx[-1]]
        if defl_perp:
            angle_label += f", Defl. perp.: {defl_perp}°"
        ax.text(
            label_pos[0],
            label_pos[1],
            angle_label,
            fontsize=8,
            ha="center",
            va="bottom",
            rotation=0,
            color="black",
        )


def plot_kz_cut(
    hv=21.2,
    Eb=0,
    polar_or_tilt=0,
    theta_par_range=(-15, 15),
    V0=12,
    ax=None,
    label_cut=True,
    **kwargs,
):
    """Plot the kz-dep of the cut along the analyser slit for given parameters.

    Parameters
    ----------
    hv : float, optional
        Photon energy (eV) (default=21.2 eV).
    Eb : float, optional
        Binding energy (eV) (positive for below the Fermi level, default=0).
    theta_par_range : tuple, optional
        Range of theta_par values to calculate over in the form (start, stop);
        default=(-15, 15)).
    polar_or_tilt : float, optional
        Angle offset in direction along the slit (deg) (default=0).
    V0 : float, optional
        Inner potential (eV) (default=12).
    ax : matplotlib.axes.Axes, optional
        Matplotlib axes object to plot on (default=None).
    label_cut : bool, optional
        Whether to label the cut with the photon energy used (default=True).
    **kwargs
        Additional keyword arguments to pass to the plot.
    """

    from peaks.core.process.k_conversion import get_kz_cut

    kx, kz = get_kz_cut(
        hv=hv,
        Eb=Eb,
        theta_par_range=theta_par_range,
        polar_or_tilt=polar_or_tilt,
        V0=V0,
    )

    if not ax:
        fig, ax = plt.subplots()

    ax.plot(kx, kz, **kwargs)

    if label_cut:
        # Add label to the line showing the photon energy
        label = f"hv: {hv} eV  "
        ax.text(
            kx[0],
            kz[0],
            label,
            fontsize=10,
            ha="right",
            va="bottom",
            rotation=0,
            color="black",
        )


def plot_nanofocus(data, focus="defocus"):
    """Function to determine the focus of a scan obtained at the nano branch of the I05 beamline at Diamond Light
    Source, and plot the results. The function works by determining the focal position at which at scanned feature
    becomes sharpest.

    Parameters
    ------------
    data : xarray.DataArray
        The 2D (or 4D) focussing scan data with some spatial direction dimension and some focussing direction
        dimension (and additionally theta_par and eV dimensions for 4D data).

    focus : str, optional
        The dimension of the data that represents the focussing direction.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        focus_scan = load('i05-1-12345.nxs')

        # Plot the results of how the sharpness of a feature varies with defocus, determining the focal position
        focus_scan.plot_nanofocus()

    """

    # Ensure the data is a 2D DataArray by integrating in energy and angle space if required
    if len(data.dims) == 4:
        data = data.mean(["theta_par", "eV"], keep_attrs=True)

    # Extract the spatial dimension
    data_dims = list(data.dims)
    data_dims.remove(focus)
    spatial = data_dims[0]

    # Determine the step of the focus dimension
    focus_step = data[focus].data[1] - data[focus].data[0]

    # Set up a subplot to plot the focussing analysis
    fig, axes = plt.subplots(figsize=(12, 14), ncols=2, nrows=3)

    # Plot the intensity map
    data.plot(x=focus, ax=axes[0, 0], add_colorbar=False)

    # Plot focus-dependent cuts through the spatial dimension (using the plot_DCs function)
    try:
        plot_DCs(data, color="black", offset=0.1, norm=False, ax=axes[0, 1], y=spatial)
    except ValueError:
        plot_DCs(
            data,
            color="black",
            offset=0.1,
            norm=False,
            ax=axes[0, 1],
            stack_dim=spatial,
        )
    axes[0, 1].set_xlabel("")
    axes[0, 1].set_xticks([])
    axes[0, 1].set_title("Line cuts", fontsize=16)

    # Calculate and plot the focus-dependent means of the absolute values of the derivatives along the spatial
    # dimension
    mean_abs_deriv = abs(data.diff(spatial)).mean(spatial)
    mean_abs_deriv.plot(ax=axes[1, 0], c="black")

    # Estimate and plot the focal point from the focus-dependent means of the absolute values of the derivatives
    # along the spatial dimension
    mean_abs_deriv_smoothed = mean_abs_deriv.smooth(**{focus: 2 * focus_step})
    mean_abs_deriv_smoothed.plot(ax=axes[1, 0], c="black", alpha=0.2)
    max_mean_abs_deriv = mean_abs_deriv_smoothed.argmax(dim="defocus")
    focal_point_mean_abs_deriv = mean_abs_deriv[focus].data[max_mean_abs_deriv]
    axes[1, 0].set_title(
        "Mean of the abs(deriv) estimate: {focal_point}".format(
            focal_point=focal_point_mean_abs_deriv
        ),
        fontsize=16,
    )
    axes[1, 0].axvline(focal_point_mean_abs_deriv, c="black", linestyle="--")
    axes[1, 0].set_ylabel("Intensity (arb. units)")
    axes[1, 0].set_yticks([])

    # Calculate and plot the focus-dependent maximums of the absolute values of the derivatives along the spatial
    # dimension
    max_abs_deriv = abs(data.diff(spatial)).max(spatial)
    max_abs_deriv.plot(ax=axes[1, 1], c="black")

    # Estimate and plot the focal point from the focus-dependent maximums of the absolute values of the derivatives
    # along the spatial dimension
    max_abs_deriv_smoothed = max_abs_deriv.smooth(**{focus: 2 * focus_step})
    max_abs_deriv_smoothed.plot(ax=axes[1, 1], c="black", alpha=0.2)
    max_max_abs_deriv = max_abs_deriv_smoothed.argmax(dim="defocus")
    focal_point_max_abs_deriv = max_abs_deriv[focus].data[max_max_abs_deriv]
    axes[1, 1].set_title(
        "Max of the abs(deriv) estimate: {focal_point}".format(
            focal_point=focal_point_max_abs_deriv
        ),
        fontsize=16,
    )
    axes[1, 1].axvline(focal_point_max_abs_deriv, c="black", linestyle="--")
    axes[1, 1].set_ylabel("Intensity (arb. units)")
    axes[1, 1].set_yticks([])

    # Calculate and plot the focus-dependent variance along the spatial dimension
    variance = data.var(spatial)
    variance.plot(ax=axes[2, 0], c="black")

    # Estimate and plot the focal point from the focus-dependent variance along the spatial dimension
    variance_smoothed = variance.smooth(**{focus: 2 * focus_step})
    variance_smoothed.plot(ax=axes[2, 0], c="black", alpha=0.2)
    max_variance = variance_smoothed.argmax(dim="defocus")
    focal_point_max_variance = variance[focus].data[max_variance]
    axes[2, 0].set_title(
        "Max of the variance estimate: {focal_point}".format(
            focal_point=focal_point_max_variance
        ),
        fontsize=16,
    )
    axes[2, 0].axvline(focal_point_max_variance, c="black", linestyle="--")
    axes[2, 0].set_ylabel("Intensity (arb. units)")
    axes[2, 0].set_yticks([])

    # Perform a fast Fourier transform (FFT) analysis of the data to look for increased higher frequencies
    # indicative of a sharper step, by transforming the spatial dimension of the data to frequency space
    FFT_data = data.copy(deep="True")
    if focus != data.dims[0]:  # Ensure data shape is as expected
        FFT_data = FFT_data.T
        FFT_data.data = abs(fft(data.pint.dequantify().T.data))  # Perform FFT
    else:
        FFT_data.data = abs(fft(data.pint.dequantify().data))  # Perform FFT
    FFT_data = FFT_data.pint.quantify(data.pint.units)
    FFT_data_out = FFT_data.isel({spatial: FFT_data.argmax(dim=spatial).data[0] + 1})

    # Plot the FFT analysis results
    FFT_data_out.plot(ax=axes[2, 1], c="black")

    # Estimate and plot the focal point from the FFT analysis
    FFT_data_out_smoothed = FFT_data_out.smooth(**{focus: 2 * focus_step})
    FFT_data_out_smoothed.plot(ax=axes[2, 1], c="black", alpha=0.2)
    max_FFT = FFT_data_out_smoothed.argmax(dim="defocus")
    focal_point_max_FFT = FFT_data_out[focus].data[max_FFT]
    axes[2, 1].set_title(
        "Fast Fourier transform estimate: {focal_point}".format(
            focal_point=focal_point_max_FFT
        ),
        fontsize=16,
    )
    axes[2, 1].axvline(focal_point_max_FFT, c="black", linestyle="--")
    axes[2, 1].set_ylabel("Intensity (arb. units)")
    axes[2, 1].set_yticks([])

    # Determine the average focal position estimate (removing any estimate outliers that are more than 2 std away
    # from the mean of the estimates)
    estimates = np.array(
        [
            focal_point_mean_abs_deriv,
            focal_point_max_abs_deriv,
            focal_point_max_variance,
            focal_point_max_FFT,
        ]
    )
    estimates_no_outliers = []
    for i, item in enumerate(abs(abs(estimates) - abs(estimates.mean()))):
        if item <= (2 * estimates.std()):
            estimates_no_outliers.append(estimates[i])
    avg_estimate = round(np.mean(estimates_no_outliers), 2)

    # Plot the average focal position
    axes[0, 0].set_title(
        "Intensity map ({focus} estimate: {avg_estimate})".format(
            focus=focus, avg_estimate=avg_estimate
        ),
        fontsize=16,
    )
    axes[0, 0].axvline(avg_estimate, c="#ECDE00", linestyle="dotted")
    axes[1, 0].axvline(avg_estimate, c="#ECDE00", linestyle="dotted")
    axes[1, 1].axvline(avg_estimate, c="#ECDE00", linestyle="dotted")
    axes[2, 0].axvline(avg_estimate, c="#ECDE00", linestyle="dotted")
    axes[2, 1].axvline(avg_estimate, c="#ECDE00", linestyle="dotted")

    plt.tight_layout()



"""Helper functions for acting on metadata."""

import copy
import inspect
from datetime import datetime
from pprint import pprint
from typing import List, Union

import pandas as pd
import xarray as xr
from IPython.display import display
from pydantic import BaseModel, Field

from peaks import __version__


# Define the structure of a single analysis history record
class AnalysisHistoryRecord(BaseModel):
    time: datetime
    peaks_version: str = Field(..., alias="peaks version")
    record: Union[str, dict]
    fn_name: str = Field(..., alias="function name")

    class Config:
        populate_by_name = (
            True  # Allow aliases for both serialization and deserialization
        )


# A model to hold the list of records
class AnalysisHistoryRecordCollection(BaseModel):
    records: List[AnalysisHistoryRecord] = []

    # Method to append new records
    def add_record(self, new_record: dict):
        record = AnalysisHistoryRecord(**new_record)
        self.records.append(record)


def _update_hist(data, record_text, fn_name=None, update_in_place=True):
    """Updates the analysis history metadata of the supplied DataArray. Can be used as a standalone function
    or via the :class:`xarray.DataArray` accessors `.history.add()` to update the history of the passed
    :class:`xarray.DataArray` in place or `history.assign()` to return a copy of the :class:`xarray.DataArray` with
    the updated history.

    Parameters
    ------------
    data : xarray.DataArray
        The :class:`xarray.DataArray` for which the analysis history is to be updated.

    record_text : string
        Descriptive string to append to the DataArray's history metadata record.

    fn_name : string, optional
        An optional function name to include in the history record. If not specified, the function name of the caller
        function will be attempted to be automatically parsed.

    update_in_place : bool, optional
        If True, updates the analysis history metadata in place. Defaults to True.


    Returns
    ------------
    xarray.DataArray, optional
        The DataArray with the updated analysis history metadata. If `update_in_place` is set `True`, returns None.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        # Load some data
        disp = pks.load('disp.ibw')

        # Update the analysis history metadata of the dispersion, modifiying the existing array in place
        disp.history.add('From inspection, seems the counts are a factor of 2 too high', update_in_place=True)
        # Display the current history metadata, which should now include the new record
        disp.history()

        disp = disp/2
        # Update the analysis history metadata of the dispersion
        disp2 = disp.history.add('Dispersion data divided by 2')

        # The history metadata of the original dispersion should not have been modified
        disp.history()
        # But the history metadata of the new dispersion should include the new record
        disp2.history()

        # Use within a function to update the history metadata, including a reference to the called function
        from peaks.utils.metadata import _update_hist
        def add_one(data):
            data += 1
            _update_hist(data, 'Data incremented by 1', fn_name='add_one', update_in_place=True)
            return data
    """

    # Get current analysis history metadata list, creating if it doesn't exist
    if "_analysis_history" not in data.attrs:
        data.attrs["_analysis_history"] = AnalysisHistoryRecordCollection()
    analysis_history = data.attrs.get("_analysis_history")

    # If updating in place, modify the existing history list, otherwise make a deep copy to avoid in-place modification
    if not update_in_place:
        analysis_history = copy.deepcopy(analysis_history)

    # If not provided, try to automatically parse the function name of the caller
    if fn_name is None:
        try:
            fn_name = inspect.stack()[1].function
        except IndexError:
            fn_name = ""

    # Format the new history entry as a JSON-style payload with additional metadata and add to the existing history
    new_record = {
        "time": datetime.now().isoformat(),
        "peaks version": str(__version__),
        "record": record_text,
        "function name": fn_name,
    }

    analysis_history.add_record(new_record)

    if not update_in_place:
        # Return the DataArray with the updated analysis history metadata,
        # using assign_attrs to avoid in-place modification
        return data.assign_attrs(_analysis_history=analysis_history)


def update_history_decorator(record_text):
    """Convenience decorator to make a function automatically add a fixed string to the analysis metadata record of an
    :class:`xarray.DataArray`.

    This decorator assumes that the first argument of the function is the :class:`xarray.DataArray` object on which
    the function is to act, and that the function returns a modified :class:`xarray.DataArray` object.

    Parameters
    ------------
    record_text : string
        Descriptive string to append to the DataArray's history metadata record.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks
        from peaks.utils.metadata import update_history_decorator

        @update_history_decorator('Data incremented by 1')
        def add_one(data):
            data += 1
            return data
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            if not isinstance(args[0], xr.DataArray):
                raise TypeError(
                    "Update history decorator assumes the first argument of the funciton is "
                    "the xarray.DataArray object on which the function is to act."
                )
            # Update the history, including passing the function name
            updated_dataarray = args[0].history.assign(
                record_text, fn_name=func.__name__
            )
            return func(updated_dataarray, *args[1:], **kwargs)

        return wrapper

    return decorator


@xr.register_dataset_accessor("history")
@xr.register_dataarray_accessor("history")
class History:
    """Custom accessor for the analysis history metadata of a :class:`xarray.DataArray`.

    This class provides methods to interact with the analysis history metadata of an `xarray.DataArray`.
    Access via the `.history` method.

    Methods
    -------
    __call__(return_history=False)
        Displays (and optionally returns) the history metadata of the :class:`xarray.DataArray`.
    add(record_text, fn_name=None)
        Updates the analysis history metadata of the :class:`xarray.DataArray`.
    """

    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def __repr__(self):
        """Make a nice in-line display of the history metadata and return a timestamp string."""
        analysis_history = self._obj.attrs.get("_analysis_history")
        if analysis_history:
            analysis_history = analysis_history.dict(by_alias=True)["records"]
        df = pd.DataFrame(analysis_history)
        with pd.option_context(
            "display.max_colwidth", None, "display.colheader_justify", "left"
        ):
            display(df.fillna(""))
        return f"History as of {datetime.now().strftime('%y-%m-%d %H-%M-%S')}"

    def __call__(self, index=None):
        """
        Return the history metadata of the DataArray.

        Returns
        -------
        list of dict
            List of the current history metadata entries

        index : int, optional
            The index of the history metadata item to display. Defaults to most recent entry.


        Examples
        --------
        Example usage is as follows::
            import peaks as pks

            disp = pks.load('disp.ibw')
            disp_k = disp.k_convert()  # Convert the data to k-space

            # Display the current history metadata
            disp_k.history()

        """

        record = (
            self._obj.attrs.get("_analysis_history")
            .records[index if index is not None else -1]
            .dict(by_alias=True)
            .copy()
        )
        pprint(record)

    def add(self, record_text, fn_name=None):
        """Update the analysis history metadata of the :class:`xarray.DataArray` in place.

        See Also
        --------
        _update_hist : The underlying function that updates the history metadata of the DataArray. This is called
        with `update_in_place=True` to update the history metadata in place.

        assign : Alternative method to update the history metadata in a copy of the DataArray.
        """
        # If not provided, try to automatically parse the function name of the caller
        if fn_name is None:
            try:
                fn_name = inspect.stack()[1].function
            except IndexError:
                fn_name = ""
        return _update_hist(self._obj, record_text, fn_name, True)

    def assign(self, record_text, fn_name=None):
        """Assign a new history item to a copy of self.

        See Also
        --------
        _update_hist : The underlying function that updates the history metadata of the DataArray. This is called
        with `update_in_place=False` to return a copy of the DataArray with the updated history metadata.

        add : Alternative method to update the history metadata in place.
        """
        # If not provided, try to automatically parse the function name of the caller
        if fn_name is None:
            try:
                fn_name = inspect.stack()[1].function
            except IndexError:
                fn_name = ""
        return _update_hist(self._obj, record_text, fn_name, False)

    def get(self, index=None, return_history=False):
        """Return all of or an item of the history metadata of the DataArray.

        Parameters
        ----------
        index : int, optional
            The index of the history metadata item to return. Defaults to returning a list of all entries.

        Returns
        -------
        list or dict
            The relevant history metadata item(s)
        """

        return (
            self._obj.attrs.get("_analysis_history")
            .dict(by_alias=True)
            .copy()["records"][index if index is not None else slice(None)]
        )

    def json(self):
        """Return the history metadata as a JSON-formatted string.

        Returns
        -------
        str
            The history metadata as a JSON-formatted string.

        Examples
        --------
        Example usage is as follows::

            import peaks as pks

            disp = load('disp.ibw')
            disp_k = disp.k_convert()

            # Get the history metadata as a JSON-formatted string
            disp_k.history.json()
        """
        return self._obj.attrs.get("_analysis_history").json(by_alias=True)

    def save(self, fname):
        """Save the history metadata as a JSON-formatted string.

        Parameters
        ----------
        fname : str
            The filename to save the history metadata to.

        Examples
        --------
        Example usage is as follows::

            import peaks as pks

            disp = load('disp.ibw')
            disp_k = disp.k_convert()

            # Save the history metadata to a JSON file
            disp_k.history.save('disp_k_history.json')
        """
        with open(fname, "w") as f:
            f.write(self.json())

    add.__doc__ = _update_hist.__doc__
